# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import platform
import socket
import sys
import time

from . import resource
from .actors import FunctionActor
from .compat import six
from .utils import git_info

try:
    import numpy as np
except ImportError:
    np = None
try:
    import scipy
except ImportError:
    scipy = None
try:
    import cupy as cp
except ImportError:
    cp = None

logger = logging.getLogger(__name__)
_collectors = dict()


def register_collector(collector):
    _collectors[id(collector)] = collector


def gather_node_info():
    from .lib.mkl_interface import mkl_get_version
    mem_stats = resource.virtual_memory()

    node_info = {
        'command_line': ' '.join(sys.argv),
        'platform': platform.platform(),
        'host_name': socket.gethostname(),
        'sys_version': sys.version,
        'cpu_used': resource.cpu_percent() / 100.0,
        'cpu_total': resource.cpu_count(),
        'memory_used': mem_stats.used,
        'memory_total': mem_stats.total,
        'update_time': time.time(),
    }

    for collector in _collectors.values():
        node_info.update(collector())

    if np is None:
        node_info['numpy_info'] = 'Not installed'
    else:
        sio = six.StringIO()
        sio.write('Version: %s\n' % np.__version__)
        if hasattr(np, '__mkl_version__') and mkl_get_version:
            mkl_version = mkl_get_version()
            sio.write('MKL Version: %d.%d.%d\n' % (mkl_version.major, mkl_version.minor, mkl_version.update))
        node_info['numpy_info'] = sio.getvalue().strip()

    if scipy is None:
        node_info['scipy_info'] = 'Not installed'
    else:
        node_info['scipy_info'] = 'Version: %s' % scipy.__version__

    git = git_info()
    if git:
        node_info['git_info'] = '%s %s' % (git[0], git[1])
    else:
        node_info['git_info'] = 'Not available'

    return node_info


class NodeInfoActor(FunctionActor):
    def __init__(self):
        super(NodeInfoActor, self).__init__()
        self._node_info = None

    @classmethod
    def default_name(cls):
        return 's:' + cls.__name__

    def post_create(self):
        logger.debug('Actor %s running in process %d', self.uid, os.getpid())

        self.ref().gather_info()

    def gather_info(self):
        self._node_info = gather_node_info()
        self.ref().gather_info(_tell=True, _delay=1)

    def get_info(self):
        return self._node_info



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .core import TensorArgReduction


class TensorNanArgminChunk(operands.NanArgminChunk, TensorArgReduction):
    def __init__(self, axis=None, dtype=np.dtype(int), keepdims=None,
                 combine_size=None, offset=None, total_shape=None,**kw):
        super(TensorNanArgminChunk, self).__init__(_axis=axis, _dtype=dtype,
                                                   _keepdims=keepdims, _combine_size=combine_size,
                                                   _offset=offset, _total_shape=total_shape, **kw)


class TensorNanArgminCombine(operands.NanArgminCombine, TensorArgReduction):
    def __init__(self, axis=None, dtype=np.dtype(int), keepdims=None, combine_size=None, **kw):
        super(TensorNanArgminCombine, self).__init__(_axis=axis, _dtype=dtype, _keepdims=keepdims,
                                                     _combine_size=combine_size, **kw)


class TensorNanArgmin(operands.NanArgmin, TensorArgReduction):
    def __init__(self, axis=None, dtype=np.dtype(int), keepdims=None, combine_size=None, **kw):
        super(TensorNanArgmin, self).__init__(_axis=axis, _dtype=dtype, _keepdims=keepdims,
                                              _combine_size=combine_size, **kw)

    @staticmethod
    def _get_op_types():
        return TensorNanArgminChunk, TensorNanArgmin, TensorNanArgminCombine


def nanargmin(a, axis=None, out=None, keepdims=None, combine_size=None):
    """
    Return the indices of the minimum values in the specified axis ignoring
    NaNs. For all-NaN slices ``ValueError`` is raised. Warning: the results
    cannot be trusted if a slice contains only NaNs and Infs.

    Parameters
    ----------
    a : array_like
        Input data.
    axis : int, optional
        Axis along which to operate.  By default flattened input is used.
    combine_size: int, optional
        The number of chunks to combine.

    Returns
    -------
    index_array : Tensor
        A tensor of indices or a single index value.

    See Also
    --------
    argmin, nanargmax

    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = mt.array([[mt.nan, 4], [2, 3]])
    >>> mt.argmin(a).execute()
    0
    >>> mt.nanargmin(a).execute()
    2
    >>> mt.nanargmin(a, axis=0).execute()
    array([1, 1])
    >>> mt.nanargmin(a, axis=1).execute()
    array([1, 0])

    """
    op = TensorNanArgmin(axis=axis, dtype=np.dtype(int), keepdims=keepdims, combine_size=combine_size)
    return op(a, out=out)



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess
import os
import sys

version_info = (0, 1, 0, 'b1')
_num_index = max(idx if isinstance(v, int) else 0
                 for idx, v in enumerate(version_info))
__version__ = '.'.join(map(str, version_info[:_num_index + 1])) + \
              ''.join(version_info[_num_index + 1:])


def _get_cmd_results(pkg_root, cmd):
    proc = subprocess.Popen(cmd, cwd=pkg_root, stdout=subprocess.PIPE)
    proc.wait()
    if proc.returncode == 0:
        s = proc.stdout.read()
        if sys.version_info[0] >= 3:
            s = s.decode()
        return s


def get_git_info():
    pkg_root = os.path.dirname(os.path.abspath(__file__))
    git_root = os.path.join(os.path.dirname(pkg_root), '.git')

    if os.path.exists(git_root):
        commit_hash = _get_cmd_results(pkg_root, ['git', 'rev-parse', 'HEAD']).strip()
        if not commit_hash:
            return
        branches = _get_cmd_results(pkg_root, ['git', 'branch']).splitlines(False)
        commit_ref = None
        for l in branches:
            if not l.startswith('*'):
                continue
            striped = l[1:].strip()
            if not striped.startswith('('):
                commit_ref = striped
            else:
                _, commit_ref = striped.rsplit(' ', 1)
                commit_ref = commit_ref.rstrip(')')
        if commit_ref is None:
            return
        return commit_hash, commit_ref
    else:
        branch_file = os.path.join(pkg_root, '.git-branch')
        if not os.path.exists(branch_file):
            return
        with open(branch_file, 'r') as bf:
            bf_content = bf.read().strip()
            return bf_content.split(None, 1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .base import HasInput
from ..operands.core import Operand
from .. import opcodes as OperandDef
from ..serialize import ValueType, KeyField, StringField, TupleField, BoolField, AnyField


class QR(HasInput):
    _op_type_ = OperandDef.QR

    _input = KeyField('input')
    _method = StringField('method')

    @property
    def method(self):
        return self._method

    @property
    def output_limit(self):
        return 2


class SVD(HasInput):
    _op_type_ = OperandDef.SVD

    _input = KeyField('input')
    _method = StringField('method')

    @property
    def method(self):
        return self._method

    @property
    def output_limit(self):
        return 3


class Cholesky(HasInput):
    _op_type_ = OperandDef.CHOLESKY

    _input = KeyField('input')
    _lower = BoolField('lower')

    @property
    def lower(self):
        return self._lower


class SolveTriangular(Operand):
    _op_type_ = OperandDef.SOLVE_TRIANGULAR

    _a = KeyField('a')
    _b = KeyField('b')
    _lower = BoolField('lower')

    @property
    def a(self):
        return self._a

    @property
    def b(self):
        return self._b

    @property
    def lower(self):
        return self._lower


class LU(HasInput):
    _pb_op_type = OperandDef.LU

    _input = KeyField('input')

    @property
    def output_limit(self):
        return 3


class Norm(HasInput):
    _op_type_ = OperandDef.NORM

    _input = KeyField('input')
    _ord = AnyField('ord')
    _axis = TupleField('axis', ValueType.int32)
    _keepdims = BoolField('keepdims')

    @property
    def ord(self):
        return getattr(self, '_ord', None)

    @property
    def axis(self):
        return self._axis

    @property
    def keepdims(self):
        return self._keepdims



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numpy.linalg import LinAlgError

from .... import operands
from ..core import TensorOperandMixin
from ..datasource import tensor as astensor


def _H(chunk):
    from ..base.transpose import TensorTranspose
    from ..arithmetic.conj import TensorConj

    trans_op = TensorTranspose(dtype=chunk.dtype)
    c = trans_op.new_chunk([chunk], chunk.shape[::-1], index=chunk.index[::-1])
    conj_op = TensorConj(dtype=c.dtype)
    return conj_op.new_chunk([c, None, None], c.shape, index=c.index)


class TensorCholesky(operands.Cholesky, TensorOperandMixin):
    def __init__(self, lower=None, dtype=None, **kw):
        super(TensorCholesky, self).__init__(_lower=lower, _dtype=dtype, **kw)

    def _set_inputs(self, inputs):
        super(TensorCholesky, self)._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, a):
        return self.new_tensor([a], a.shape)

    @classmethod
    def tile(cls, op):
        from ..datasource.zeros import TensorZeros
        from ..arithmetic.subtract import TensorSubtract
        from ..arithmetic.utils import tree_add
        from .dot import TensorDot
        from .solve_triangular import TensorSolveTriangular

        tensor = op.outputs[0]
        in_tensor = op.input
        if len(set(in_tensor.nsplits[0] + in_tensor.nsplits[1])) != 1:
            nsplit = min(in_tensor.nsplits[0] + in_tensor.nsplits[1])
            # input's chunks must be all square
            in_tensor = in_tensor.rechunk([nsplit for _ in range(in_tensor.ndim)]).single_tiles()
            if len(set(in_tensor.nsplits[0] + in_tensor.nsplits[1])) != 1:
                raise LinAlgError('All chunks must be a square matrix to perform cholesky decomposition.')

        lower_chunks, upper_chunks = {}, {}
        for i in range(in_tensor.chunk_shape[0]):
            for j in range(in_tensor.chunk_shape[1]):
                if i < j:
                    lower_chunk = TensorZeros(dtype=tensor.dtype).new_chunk(
                        None, (in_tensor.nsplits[0][i], in_tensor.nsplits[1][j]), index=(i, j))
                    upper_chunk = TensorZeros(dtype=tensor.dtype).new_chunk(
                        None, (in_tensor.nsplits[0][i], in_tensor.nsplits[1][j]), index=(j, i))
                    lower_chunks[lower_chunk.index] = lower_chunk
                    upper_chunks[upper_chunk.index] = upper_chunk
                elif i == j:
                    target = in_tensor.cix[i, j]
                    if i > 0:
                        prev_chunks = []
                        for p in range(i):
                            a, b = lower_chunks[i, p], upper_chunks[p, j]
                            prev_chunk = TensorDot(dtype=tensor.dtype).new_chunk(
                                [a, b], (a.shape[0], b.shape[1]))
                            prev_chunks.append(prev_chunk)
                        if len(prev_chunks) == 1:
                            s = prev_chunks[0]
                        else:
                            s = tree_add(prev_chunks[0].dtype, prev_chunks,
                                         None, prev_chunks[0].shape)
                        target = TensorSubtract(dtype=tensor.dtype).new_chunk(
                            [target, s, None, None], target.shape)
                    lower_chunk = TensorCholesky(lower=True, dtype=tensor.dtype).new_chunk(
                        [target], target.shape, index=(i, j))
                    upper_chunk = _H(lower_chunk)
                    lower_chunks[lower_chunk.index] = lower_chunk
                    upper_chunks[upper_chunk.index] = upper_chunk
                else:
                    target = in_tensor.cix[j, i]
                    if j > 0:
                        prev_chunks = []
                        for p in range(j):
                            a, b = lower_chunks[j, p], upper_chunks[p, i]
                            prev_chunk = TensorDot(dtype=tensor.dtype).new_chunk(
                                [a, b], (a.shape[0], b.shape[1]))
                            prev_chunks.append(prev_chunk)
                        if len(prev_chunks) == 1:
                            s = prev_chunks[0]
                        else:
                            s = tree_add(prev_chunks[0].dtype, prev_chunks,
                                         None, prev_chunks[0].shape)
                        target = TensorSubtract(dtype=tensor.dtype).new_chunk(
                            [target, s, None, None], target.shape)
                    upper_chunk = TensorSolveTriangular(lower=True, dtype=tensor.dtype).new_chunk(
                        [lower_chunks[j, j], target], target.shape, index=(j, i))
                    lower_chunk = _H(upper_chunk)
                    lower_chunks[lower_chunk.index] = lower_chunk
                    upper_chunks[upper_chunk.index] = upper_chunk

        new_op = op.copy()
        if op.lower:
            return new_op.new_tensors(op.inputs, tensor.shape,
                                      chunks=list(lower_chunks.values()), nsplits=in_tensor.nsplits)
        else:
            return new_op.new_tensors(op.inputs, tensor.shape,
                                      chunks=list(upper_chunks.values()), nsplits=in_tensor.nsplits)


def cholesky(a, lower=False):
    """
    Cholesky decomposition.

    Return the Cholesky decomposition, `L * L.H`, of the square matrix `a`,
    where `L` is lower-triangular and .H is the conjugate transpose operator
    (which is the ordinary transpose if `a` is real-valued).  `a` must be
    Hermitian (symmetric if real-valued) and positive-definite.  Only `L` is
    actually returned.

    Parameters
    ----------
    a : (..., M, M) array_like
        Hermitian (symmetric if all elements are real), positive-definite
        input matrix.
    lower : bool
        Whether to compute the upper or lower triangular Cholesky
        factorization.  Default is upper-triangular.

    Returns
    -------
    L : (..., M, M) array_like
        Upper or lower-triangular Cholesky factor of `a`.

    Raises
    ------
    LinAlgError
       If the decomposition fails, for example, if `a` is not
       positive-definite.

    Notes
    -----

    Broadcasting rules apply, see the `mt.linalg` documentation for
    details.

    The Cholesky decomposition is often used as a fast way of solving

    .. math:: A \\mathbf{x} = \\mathbf{b}

    (when `A` is both Hermitian/symmetric and positive-definite).

    First, we solve for :math:`\\mathbf{y}` in

    .. math:: L \\mathbf{y} = \\mathbf{b},

    and then for :math:`\\mathbf{x}` in

    .. math:: L.H \\mathbf{x} = \\mathbf{y}.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> A = mt.array([[1,-2j],[2j,5]])
    >>> A.execute()
    array([[ 1.+0.j,  0.-2.j],
           [ 0.+2.j,  5.+0.j]])
    >>> L = mt.linalg.cholesky(A, lower=True)
    >>> L.execute()
    array([[ 1.+0.j,  0.+0.j],
           [ 0.+2.j,  1.+0.j]])
    >>> mt.dot(L, L.T.conj()).execute() # verify that L * L.H = A
    array([[ 1.+0.j,  0.-2.j],
           [ 0.+2.j,  5.+0.j]])
    >>> A = [[1,-2j],[2j,5]] # what happens if A is only array_like?
    >>> mt.linalg.cholesky(A, lower=True).execute()
    array([[ 1.+0.j,  0.+0.j],
           [ 0.+2.j,  1.+0.j]])

    """
    a = astensor(a)

    if a.ndim != 2:
        raise LinAlgError('{0}-dimensional array given. '
                          'Tensor must be two-dimensional'.format(a.ndim))
    if a.shape[0] != a.shape[1]:
        raise LinAlgError('Input must be square')

    cho = np.linalg.cholesky(np.array([[1, 2], [2, 5]], dtype=a.dtype))

    op = TensorCholesky(lower=lower, dtype=cho.dtype)
    return op(a)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from ..utils import infer_dtype
from .core import TensorUnaryOp


class TensorLog2(operands.Log2, TensorUnaryOp):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorLog2, self).__init__(_casting=casting, _err=err,
                                         _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x):
        return False


@infer_dtype(np.log2)
def log2(x, out=None, where=None, **kwargs):
    """
    Base-2 logarithm of `x`.

    Parameters
    ----------
    x : array_like
        Input values.
    out : Tensor, None, or tuple of tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    y : Tensor
        Base-2 logarithm of `x`.

    See Also
    --------
    log, log10, log1p

    Logarithm is a multivalued function: for each `x` there is an infinite
    number of `z` such that `2**z = x`. The convention is to return the `z`
    whose imaginary part lies in `[-pi, pi]`.

    For real-valued input data types, `log2` always returns real output.
    For each value that cannot be expressed as a real number or infinity,
    it yields ``nan`` and sets the `invalid` floating point error flag.

    For complex-valued input, `log2` is a complex analytical function that
    has a branch cut `[-inf, 0]` and is continuous from above on it. `log2`
    handles the floating-point negative zero as an infinitesimal negative
    number, conforming to the C99 standard.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> x = mt.array([0, 1, 2, 2**4])
    >>> mt.log2(x).execute()
    array([-Inf,   0.,   1.,   4.])

    >>> xi = mt.array([0+1.j, 1, 2+0.j, 4.j])
    >>> mt.log2(xi).execute()
    array([ 0.+2.26618007j,  0.+0.j        ,  1.+0.j        ,  2.+2.26618007j])
    """
    op = TensorLog2(**kwargs)
    return op(x, out=out, where=where)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .core import TensorRandomOperandMixin, handle_array


class TensorGamma(operands.Gamma, TensorRandomOperandMixin):
    __slots__ = '_shape', '_scale', '_size'
    _input_fields_ = ['_shape', '_scale']

    def __init__(self, state=None, size=None, dtype=None, gpu=None, **kw):
        dtype = np.dtype(dtype) if dtype is not None else dtype
        super(TensorGamma, self).__init__(_state=state, _size=size, _dtype=dtype,
                                          _gpu=gpu, **kw)

    def __call__(self, shape, scale, chunk_size=None):
        return self.new_tensor([shape, scale], None, raw_chunk_size=chunk_size)


def gamma(random_state, shape, scale=1.0, size=None, chunk_size=None, gpu=None, **kw):
    r"""
    Draw samples from a Gamma distribution.

    Samples are drawn from a Gamma distribution with specified parameters,
    `shape` (sometimes designated "k") and `scale` (sometimes designated
    "theta"), where both parameters are > 0.

    Parameters
    ----------
    shape : float or array_like of floats
        The shape of the gamma distribution. Should be greater than zero.
    scale : float or array_like of floats, optional
        The scale of the gamma distribution. Should be greater than zero.
        Default is equal to 1.
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  If size is ``None`` (default),
        a single value is returned if ``shape`` and ``scale`` are both scalars.
        Otherwise, ``np.broadcast(shape, scale).size`` samples are drawn.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    -------
    out : Tensor or scalar
        Drawn samples from the parameterized gamma distribution.

    See Also
    --------
    scipy.stats.gamma : probability density function, distribution or
        cumulative density function, etc.

    Notes
    -----
    The probability density for the Gamma distribution is

    .. math:: p(x) = x^{k-1}\frac{e^{-x/\theta}}{\theta^k\Gamma(k)},

    where :math:`k` is the shape and :math:`\theta` the scale,
    and :math:`\Gamma` is the Gamma function.

    The Gamma distribution is often used to model the times to failure of
    electronic components, and arises naturally in processes for which the
    waiting times between Poisson distributed events are relevant.

    References
    ----------
    .. [1] Weisstein, Eric W. "Gamma Distribution." From MathWorld--A
           Wolfram Web Resource.
           http://mathworld.wolfram.com/GammaDistribution.html
    .. [2] Wikipedia, "Gamma distribution",
           http://en.wikipedia.org/wiki/Gamma_distribution

    Examples
    --------
    Draw samples from the distribution:

    >>> import mars.tensor as mt

    >>> shape, scale = 2., 2.  # mean=4, std=2*sqrt(2)
    >>> s = mt.random.gamma(shape, scale, 1000).execute()

    Display the histogram of the samples, along with
    the probability density function:

    >>> import matplotlib.pyplot as plt
    >>> import scipy.special as sps
    >>> import numpy as np
    >>> count, bins, ignored = plt.hist(s, 50, normed=True)
    >>> y = bins**(shape-1)*(np.exp(-bins/scale) /
    ...                      (sps.gamma(shape)*scale**shape))
    >>> plt.plot(bins, y, linewidth=2, color='r')
    >>> plt.show()
    """
    if 'dtype' not in kw:
        kw['dtype'] = np.random.RandomState().gamma(
            handle_array(shape), handle_array(scale), size=(0,)).dtype
    size = random_state._handle_size(size)
    op = TensorGamma(state=random_state._state, size=size, gpu=gpu, **kw)
    return op(shape, scale, chunk_size=chunk_size)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .core import issparse, get_array_module, is_cupy, cp, cps, \
    get_sparse_module, naked, sps
from .array import SparseNDArray


def zeros_sparse_matrix(shape, dtype=float, gpu=False):
    m = sps if not gpu else cps
    return SparseMatrix(m.csr_matrix(shape, dtype=np.dtype(dtype)))


def diag_sparse_matrix(v, k=0, gpu=False):
    v = naked(v)
    if gpu and get_array_module(v) is not cp:
        v = cp.asarray(v)
    if not gpu and get_array_module(v) is not np:
        v = v.get()

    sparse_m = sps if not gpu else cps
    m = n = v.size + k
    mat = sparse_m.spdiags(v[None], [k], m, n, format='csr')
    return SparseMatrix(mat)


def eye_sparse_matrix(N, M=None, k=0, dtype=float, gpu=False):
    m = sps if not gpu else cps
    return SparseMatrix(m.eye(N, n=M, k=k, dtype=dtype, format='csr'))


def triu_sparse_matrix(m, k=0, gpu=False):
    m = naked(m)
    if gpu and get_array_module(m) is not cp:
        m = cp.asarray(m)
    if not gpu and get_array_module(m) is not np:
        m = m.get()

    sparse_m = sps if not gpu else cps
    mat = sparse_m.triu(m, k=k)
    return SparseMatrix(mat)


def tril_sparse_matrix(m, k=0, gpu=False):
    m = naked(m)
    if gpu and get_array_module(m) is not cp:
        m = cp.asarray(m)
    if not gpu and get_array_module(m) is not np:
        m = m.get()

    sparse_m = sps if not gpu else cps
    mat = sparse_m.tril(m, k=k)
    return SparseMatrix(mat)


def where(cond, x, y):
    cond, x, y = [SparseMatrix(i) if issparse(i) else i
                  for i in (cond, x, y)]
    return cond * x + (cond * (-y) + y)


class SparseMatrix(SparseNDArray):
    __slots__ = 'spmatrix',

    def __init__(self, spmatrix):
        if isinstance(spmatrix, SparseMatrix):
            self.spmatrix = spmatrix.spmatrix
        else:
            self.spmatrix = spmatrix.tocsr()

    @property
    def ndim(self):
        return self.spmatrix.ndim

    def toarray(self):
        return self.spmatrix.toarray()

    def todense(self):
        return self.spmatrix.toarray()

    def ascupy(self):
        is_cp = get_array_module(self.spmatrix) is cp
        if is_cp:
            return self
        mat_tuple = (cp.asarray(self.data), cp.asarray(self.indices), cp.asarray(self.indptr))
        return SparseMatrix(cps.csr_matrix(mat_tuple, shape=self.shape))

    def asscipy(self):
        is_cp = get_array_module(self.spmatrix) is cp
        if not is_cp:
            return self
        return SparseMatrix(self.spmatrix.get())

    def __array__(self, dtype=None):
        x = self.toarray()
        if dtype and x.dtype != dtype:
            return x.astype(dtype)
        return x

    @property
    def nbytes(self):
        return self.spmatrix.data.nbytes + self.spmatrix.indptr.nbytes \
               + self.spmatrix.indices.nbytes

    @property
    def raw(self):
        return self.spmatrix

    @property
    def data(self):
        return self.spmatrix.data

    @property
    def indptr(self):
        return self.spmatrix.indptr

    @property
    def indices(self):
        return self.spmatrix.indices

    @property
    def nnz(self):
        return self.spmatrix.nnz

    @property
    def shape(self):
        return self.spmatrix.shape

    @property
    def dtype(self):
        return self.spmatrix.dtype

    def copy(self):
        return SparseMatrix(self.spmatrix.copy())

    @property
    def real(self):
        xps = get_sparse_module(self.spmatrix)
        return SparseMatrix(xps.csr_matrix(
            (self.spmatrix.data.real, self.spmatrix.indices, self.spmatrix.indptr),
            self.spmatrix.shape
        ))

    @real.setter
    def real(self, r):
        xps = get_sparse_module(self.spmatrix)
        x = self.spmatrix.toarray()
        if issparse(r):
            r = r.toarray()
        x.real = r
        self.spmatrix = xps.csr_matrix(x)

    @property
    def imag(self):
        xps = get_sparse_module(self.spmatrix)
        return SparseMatrix(xps.csr_matrix(
            (self.spmatrix.data.imag, self.spmatrix.indices, self.spmatrix.indptr),
            self.spmatrix.shape
        ))

    @imag.setter
    def imag(self, imag):
        xps = get_sparse_module(self.spmatrix)
        x = self.spmatrix.toarray()
        if issparse(imag):
            imag = imag.toarray()
        x.imag = imag
        self.spmatrix = xps.csr_matrix(x)

    def __getattr__(self, attr):
        is_cp = get_array_module(self.spmatrix) is cp
        if attr == 'device' and is_cp:
            try:
                return self.spmatrix.device
            except NotImplementedError:
                return cp.cuda.Device(0)
        if attr == 'get' and is_cp:
            return lambda: SparseMatrix(self.spmatrix.get())

        return super(SparseMatrix, self).__getattribute__(attr)

    def astype(self, dtype):
        dtype = np.dtype(dtype)
        if self.dtype == dtype:
            return self
        return SparseMatrix(self.spmatrix.astype(dtype))

    def transpose(self, axes=None):
        assert axes is None or tuple(axes) == (1, 0)
        return SparseMatrix(self.spmatrix.transpose())

    def swapaxes(self, axis1, axis2):
        if axis1 == 0 and axis2 == 1:
            return self

        assert axis1 == 1 and axis2 == 0
        return self.transpose()

    def reshape(self, shape):
        return SparseMatrix(self.spmatrix.tolil().reshape(shape))

    def broadcast_to(self, shape):
        # TODO(jisheng): implement broadcast_to
        raise NotImplementedError

    def squeeze(self, axis=None):
        # TODO(jisheng): implement squeeze
        raise NotImplementedError

    @property
    def T(self):
        return SparseMatrix(self.spmatrix.T)

    # ---------------- arithmetic ----------------------

    def __add__(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented
        xp = get_array_module(self.spmatrix)
        if xp is cp and issparse(other) and \
                xp.all(self.spmatrix.indices == other.indices) and \
                xp.all(self.spmatrix.indptr == other.indptr):
            x = cps.csr_matrix(
                (self.spmatrix.data + other.data, self.spmatrix.indices, self.spmatrix.indptr),
                self.spmatrix.shape)
        else:
            try:
                x = self.spmatrix + other
            except NotImplementedError:
                x = self.spmatrix.toarray() + other
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def __radd__(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented
        xp = get_array_module(self.spmatrix)
        if xp is cp and issparse(other) and \
                xp.all(self.spmatrix.indices == other.indices) and \
                xp.all(self.spmatrix.indptr == other.indptr):
            x = cps.csr_matrix(
                (other.data + self.spmatrix.data, self.spmatrix.indices, self.spmatrix.indptr),
                self.spmatrix.shape)
        else:
            try:
                x = other + self.spmatrix
            except NotImplementedError:
                x = other + self.spmatrix.toarray()
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def __sub__(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented
        try:
            x = self.spmatrix - other
        except NotImplementedError:
            x = self.spmatrix.toarray() - other
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def __rsub__(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented
        try:
            x = other - self.spmatrix
        except NotImplementedError:
            x = other - self.spmatrix.toarray()
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def __mul__(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented
        if is_cupy(self.spmatrix):
            if not cp.isscalar(other):
                # TODO(jisheng): cupy does not implement multiply method
                is_other_sparse = issparse(other)
                if is_other_sparse and self.spmatrix.nnz == other.nnz and \
                        cp.all(self.spmatrix.indptr == other.indptr) and \
                        cp.all(self.spmatrix.indices == other.indices):
                    x = cps.csr_matrix((self.spmatrix.data * other.data,
                                        self.spmatrix.indices,
                                        self.spmatrix.indptr), self.spmatrix.shape)
                else:
                    if is_other_sparse:
                        other = other.toarray()
                    dense = self.spmatrix.toarray()
                    res = cp.multiply(dense, other, out=dense)
                    x = cps.csr_matrix(res)
            else:
                x = self.spmatrix * other
        else:
            x = self.spmatrix.multiply(other)
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def __rmul__(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented
        if is_cupy(self.spmatrix):
            if not cp.isscalar(other):
                # TODO(jisheng): cupy does not implement multiply method
                is_other_sparse = issparse(other)
                if is_other_sparse and self.spmatrix.nnz == other.nnz and \
                        cp.all(self.spmatrix.indptr == other.indptr) and \
                        cp.all(self.spmatrix.indices == other.indices):
                    x = cps.csr_matrix((other.data * self.spmatrix.data,
                                        self.spmatrix.indices,
                                        self.spmatrix.indptr), self.spmatrix.shape)
                else:
                    if is_other_sparse:
                        other = other.toarray()
                    dense = self.spmatrix.toarray()
                    res = cp.multiply(other, dense, out=dense)
                    x = cps.csr_matrix(res)
            else:
                x = other * self.spmatrix
        else:
            x = self.spmatrix.multiply(other)
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def __div__(self, other):
        return self.__truediv__(other)

    def __rdiv__(self, other):
        return self.__rtruediv__(other)

    def __truediv__(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented
        x = self.spmatrix / other
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def __rtruediv__(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented
        try:
            x = other / self.spmatrix
        except TypeError:
            x = other / self.spmatrix.toarray()
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def __floordiv__(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented
        if get_array_module(other).isscalar(other):
            m = get_sparse_module(self.spmatrix)
            data = self.spmatrix.data // other
            x = m.csr_matrix((data, self.spmatrix.indices, self.spmatrix.indptr),
                             self.spmatrix.shape)
        else:
            if issparse(other):
                other = other.toarray()
            x = get_sparse_module(self.spmatrix).csr_matrix(
                self.spmatrix.toarray() // other)
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def __rfloordiv__(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented
        is_sparse = issparse(other)
        if is_sparse:
            other = other.toarray()
        x = other // self.spmatrix.toarray()
        if is_sparse:
            x = get_sparse_module(x).csr_matrix(x)
        return get_array_module(x).asarray(x)

    def __pow__(self, other, modulo=None):
        if modulo is not None:
            return NotImplemented

        try:
            other = naked(other)
        except TypeError:
            return NotImplemented
        if get_array_module(other).isscalar(other):
            if other >= 0:
                x = self.spmatrix.power(other)
            else:
                data = 1 / (self.spmatrix.data ** -other)
                x = get_sparse_module(self.spmatrix).csr_matrix(
                    (data, self.spmatrix.indices, self.spmatrix.indptr),
                    self.spmatrix.shape)
        else:
            if issparse(other):
                other = other.toarray()
            x = self.spmatrix.toarray() ** other
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def __rpow__(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented
        if issparse(other):
            other = other.toarray()
        x = other ** self.spmatrix.toarray()
        return get_array_module(x).asarray(x)

    def float_power(self, other):
        ret = self.__pow__(other)
        ret = naked(ret).astype(float)
        if issparse(ret):
            return SparseMatrix(ret)
        return ret

    def __mod__(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented
        if get_array_module(other).isscalar(other):
            data = self.spmatrix.data % other
            x = get_sparse_module(self.spmatrix).csr_matrix(
                (data, self.spmatrix.indices, self.spmatrix.indptr), self.spmatrix.shape)
        else:
            if issparse(other):
                other = other.toarray()
            x = get_sparse_module(self.spmatrix).csr_matrix(
                self.spmatrix.toarray() % other)
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def __rmod__(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented
        is_sparse = issparse(other)
        if issparse(other):
            other = other.toarray()
        if get_array_module(other).isscalar(other):
            data = other % self.spmatrix.data
            x = get_sparse_module(self.spmatrix).csr_matrix(
                (data, self.spmatrix.indices, self.spmatrix.indptr), self.spmatrix.shape)
        else:
            x = other % self.spmatrix.toarray()
            if is_sparse:
                x = get_sparse_module(self.spmatrix).csr_matrix(x)
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def fmod(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        xp = get_array_module(self.spmatrix)
        if get_array_module(other).isscalar(other):
            data = xp.fmod(self.spmatrix.data, other)
            x = get_sparse_module(self.spmatrix).csr_matrix(
                (data, self.spmatrix.indices, self.spmatrix.indptr), self.spmatrix.shape)
        else:
            if issparse(other):
                other = other.toarray()
            x = get_sparse_module(self.spmatrix).csr_matrix(
                xp.fmod(self.spmatrix.toarray(), other))
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def logaddexp(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        xp = get_array_module(self.spmatrix)
        if issparse(other):
            other = other.toarray()
        return xp.logaddexp(self.spmatrix.toarray(), other)

    def logaddexpr2(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        xp = get_array_module(self.spmatrix)
        if issparse(other):
            other = other.toarray()
        return xp.logaddexp2(self.spmatrix.toarray(), other)

    def __neg__(self):
        return SparseMatrix(-self.spmatrix)

    def __pos__(self):
        return SparseMatrix(self.spmatrix.copy())

    def __abs__(self):
        return SparseMatrix(abs(self.spmatrix))

    def fabs(self):
        xp = get_array_module(self.spmatrix)
        return SparseMatrix(get_sparse_module(self.spmatrix).csr_matrix(
            xp.abs(self.spmatrix), dtype='f8'))

    def rint(self):
        return SparseMatrix(self.spmatrix.rint())

    def sign(self):
        return SparseMatrix(self.spmatrix.sign())

    def conj(self):
        return SparseMatrix(self.spmatrix.conj())

    def exp(self):
        xp = get_array_module(self.spmatrix)
        return xp.exp(self.spmatrix.toarray())

    def exp2(self):
        xp = get_array_module(self.spmatrix)
        return xp.exp2(self.spmatrix.toarray())

    def log(self):
        xp = get_array_module(self.spmatrix)
        return xp.log(self.spmatrix.toarray())

    def log2(self):
        xp = get_array_module(self.spmatrix)
        return xp.log2(self.spmatrix.toarray())

    def log10(self):
        xp = get_array_module(self.spmatrix)
        return xp.log10(self.spmatrix.toarray())

    def expm1(self):
        return SparseMatrix(self.spmatrix.expm1())

    def log1p(self):
        return SparseMatrix(self.spmatrix.log1p())

    def sqrt(self):
        return SparseMatrix(self.spmatrix.sqrt())

    def square(self):
        xp = get_array_module(self.spmatrix)
        data = xp.square(self.spmatrix.data)
        x = get_sparse_module(self.spmatrix).csr_matrix(
            (data, self.spmatrix.indices, self.spmatrix.indptr), self.spmatrix.shape
        )
        return SparseMatrix(x)

    def cbrt(self):
        xp = get_array_module(self.spmatrix)
        if hasattr(xp, 'cbrt'):
            data = xp.cbrt(self.spmatrix.data)
        else:
            data = self.spmatrix.data ** (1 / 3)
        x = get_sparse_module(self.spmatrix).csr_matrix(
            (data, self.spmatrix.indices, self.spmatrix.indptr), self.spmatrix.shape
        )
        return SparseMatrix(x)

    def reciprocal(self):
        xp = get_array_module(self.spmatrix)
        return xp.reciprocal(self.spmatrix.toarray())

    def __eq__(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        if is_cupy(self.spmatrix):
            return NotImplemented
        else:
            x = self.spmatrix == other
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def __ne__(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        if is_cupy(self.spmatrix):
            return NotImplemented
        else:
            x = self.spmatrix != other
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def __lt__(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        if is_cupy(self.spmatrix):
            return NotImplemented
        else:
            x = self.spmatrix < other
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def __le__(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        if is_cupy(self.spmatrix):
            return NotImplemented
        else:
            x = self.spmatrix <= other
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def __gt__(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        if is_cupy(self.spmatrix):
            return NotImplemented
        else:
            x = self.spmatrix > other
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def __ge__(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        if is_cupy(self.spmatrix):
            return NotImplemented
        else:
            try:
                x = self.spmatrix >= other
            except NotImplementedError:
                x = self.spmatrix.toarray() >= other
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def logical_and(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        if is_cupy(self.spmatrix):
            return NotImplemented
        else:
            other_xp = get_array_module(other)
            if other_xp.isscalar(other):
                other = other_xp.array(other).astype(bool)
            else:
                other = other.astype(bool)
            x = self.spmatrix.astype(bool).multiply(other)
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def logical_or(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        if is_cupy(self.spmatrix):
            return NotImplemented
        else:
            other_xp = get_array_module(other)
            if other_xp.isscalar(other):
                if other != 0:
                    x = np.logical_and(self.spmatrix.toarray(), other)
                else:
                    x = self.spmatrix.astype(bool)
            else:
                other = other.astype(bool)
                x = (self.spmatrix.astype(bool) + other).astype(bool)
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def logical_xor(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        if is_cupy(self.spmatrix):
            return NotImplemented
        else:
            other_xp = get_array_module(other)
            if other_xp.isscalar(other):
                other = other_xp.array(other).astype(bool)
            else:
                other = other.astype(bool)
            x = self.spmatrix.astype(bool) != other
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def logical_not(self):
        xp = get_array_module(self.spmatrix)
        return xp.logical_not(self.spmatrix.toarray())

    @staticmethod
    def _bitwise(this, other, method_name):
        try:
            this = naked(this)
        except TypeError:
            return NotImplemented
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        if not issparse(this):
            return SparseMatrix._bitwise(other, this, method_name)

        if issparse(other):
            other = other.toarray()

        xp = get_array_module(this)
        xps = get_sparse_module(this)
        return SparseMatrix(xps.csr_matrix(getattr(xp, method_name)(this.toarray(), other)))

    def __and__(self, other):
        return self._bitwise(self.spmatrix, other, 'bitwise_and')

    def __rand__(self, other):
        return self._bitwise(other, self.spmatrix, 'bitwise_and')

    def __or__(self, other):
        return self._bitwise(self.spmatrix, other, 'bitwise_or')

    def __ror__(self, other):
        return self._bitwise(other, self.spmatrix, 'bitwise_or')

    def __xor__(self, other):
        return self._bitwise(self.spmatrix, other, 'bitwise_xor')

    def __rxor__(self, other):
        return self._bitwise(other, self.spmatrix, 'bitwise_xor')

    def isclose(self, other, **kw):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        xp = get_array_module(other)
        if issparse(other):
            other = other.toarray()
        return xp.isclose(self.spmatrix.toarray(), other, **kw)

    def __invert__(self):
        xp = get_array_module(self.spmatrix)
        return xp.invert(self.spmatrix.toarray())

    @staticmethod
    def _shift(this, other, method_name):
        try:
            this = naked(this)
        except TypeError:
            return NotImplemented
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        xps = get_sparse_module(this)
        xp = get_array_module(this)

        if xp.isscalar(this):
            other_xp = get_array_module(other)
            data = getattr(other_xp, method_name)(this, other.data)
            indices, indptr, shape = other.indices, other.indptr, other.shape
        elif isinstance(this, xp.ndarray):
            # dense
            return getattr(xp, method_name)(this, other.toarray())
        else:
            tp = np.int32 if is_cupy(this) else np.bool_  # cupy.sparse does not support bool
            mask = xps.csr_matrix(
                ((this.data > 0).astype(tp), this.indices, this.indptr),
                this.shape)
            other = mask.multiply(other)
            indices, indptr, shape = this.indices, this.indptr, this.shape
            data = getattr(xp, method_name)(this.data, other.data)

        return SparseMatrix(xps.csr_matrix((data, indices, indptr), shape))

    def __lshift__(self, other):
        return self._shift(self.spmatrix, other, 'left_shift')

    def __rlshift__(self, other):
        return self._shift(other, self.spmatrix, 'left_shift')

    def __rshift__(self, other):
        return self._shift(self.spmatrix, other, 'right_shift')

    def __rrshift__(self, other):
        return self._shift(other, self.spmatrix, 'right_shift')

    def sin(self):
        return SparseMatrix(self.spmatrix.sin())

    def cos(self):
        xp = get_array_module(self.spmatrix)
        return xp.cos(self.spmatrix.toarray())

    def tan(self):
        return SparseMatrix(self.spmatrix.tan())

    def arcsin(self):
        return SparseMatrix(self.spmatrix.arcsin())

    def arccos(self):
        xp = get_array_module(self.spmatrix)
        return xp.arccos(self.spmatrix.toarray())

    def arctan(self):
        return SparseMatrix(self.spmatrix.arctan())

    def arctan2(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        xp = get_array_module(self.spmatrix)
        if issparse(other):
            other = other.toarray()
        x = xp.arctan2(self.spmatrix.toarray(), other)
        return SparseMatrix(get_sparse_module(x).csr_matrix(x))

    def hypot(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        xp = get_array_module(self.spmatrix)
        if issparse(other):
            other = other.toarray()
        x = xp.hypot(self.spmatrix.toarray(), other)
        return SparseMatrix(get_sparse_module(x).csr_matrix(x))

    def sinh(self):
        return SparseMatrix(self.spmatrix.sinh())

    def cosh(self):
        xp = get_array_module(self.spmatrix)
        return xp.cosh(self.spmatrix.toarray())

    def tanh(self):
        return SparseMatrix(self.spmatrix.tanh())

    def arcsinh(self):
        return SparseMatrix(self.spmatrix.arcsinh())

    def arccosh(self):
        xp = get_array_module(self.spmatrix)
        return xp.arccosh(self.spmatrix.toarray())

    def arctanh(self):
        return SparseMatrix(self.spmatrix.arctanh())

    def around(self, decimals=0):
        xp = get_array_module(self.spmatrix)
        data = xp.around(self.spmatrix.data, decimals=decimals)
        x = get_sparse_module(self.spmatrix).csr_matrix(
            (data, self.spmatrix.indices, self.spmatrix.indptr), self.spmatrix.shape
        )
        return SparseMatrix(x)

    def deg2rad(self):
        return SparseMatrix(self.spmatrix.deg2rad())

    def rad2deg(self):
        return SparseMatrix(self.spmatrix.rad2deg())

    def angle(self, deg=0):
        xp = get_array_module(self.spmatrix)
        data = xp.angle(self.spmatrix.data, deg=deg)
        x = get_sparse_module(self.spmatrix).csr_matrix(
            (data, self.spmatrix.indices, self.spmatrix.indptr), self.spmatrix.shape
        )
        return SparseMatrix(x)

    def dot(self, other, sparse=True):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        if sparse:
            x = self.spmatrix.dot(other)
        else:
            a = self.spmatrix.toarray()
            if issparse(other):
                other = other.toarray()
            x = a.dot(other)
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def concatenate(self, other, axis=0):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        if issparse(other):
            xps = get_sparse_module(self.spmatrix)
            if axis not in (0, 1):
                raise ValueError('axis can only be 0 or 1')
            method = xps.vstack if axis == 0 else xps.hstack
            x = method((self.spmatrix, other))
        else:
            xp = get_array_module(self.spmatrix)
            x = xp.concatenate((self.spmatrix.toarray(), other), axis=axis)

        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def _reduction(self, method_name, axis=None, dtype=None, keepdims=None, todense=False, **kw):
        if not todense:
            assert keepdims is None or keepdims is False

        if isinstance(axis, tuple):
            if sorted(axis) != [0, 1]:
                assert len(axis) == 1
                axis = axis[0]
            else:
                axis = None

        if todense:
            x = self.spmatrix.toarray()
            x = getattr(get_array_module(x), method_name)(x, axis=axis, **kw)
        else:
            x = getattr(self.spmatrix, method_name)(axis=axis, **kw)
        if issparse(x):
            return SparseMatrix(x)
        shape = (set(self.spmatrix.shape) - set([axis])).pop()
        m = get_array_module(x)
        if m.isscalar(x):
            return m.array([x])[0]
        else:
            return m.asarray(x).reshape(shape)

    def sum(self, axis=None, dtype=None, keepdims=None):
        return self._reduction('sum', axis=axis, dtype=dtype, keepdims=keepdims)

    def prod(self, axis=None, dtype=None, keepdims=None):
        return self._reduction('sum', axis=axis, dtype=dtype, keepdims=keepdims, todense=True)

    def amax(self, axis=None, dtype=None, keepdims=None):
        return self._reduction('max', axis=axis, dtype=dtype, keepdims=keepdims)

    def amin(self, axis=None, dtype=None, keepdims=None):
        return self._reduction('min', axis=axis, dtype=dtype, keepdims=keepdims)

    def all(self, axis=None, dtype=None, keepdims=None):
        ret = self._reduction('all', axis=axis, dtype=dtype, keepdims=keepdims, todense=True)
        if not issparse(ret):
            xps = get_sparse_module(self.spmatrix)
            ret = SparseMatrix(xps.csr_matrix(ret))
            return ret
        return ret

    def any(self, axis=None, dtype=None, keepdims=None):
        ret = self._reduction('any', axis=axis, dtype=dtype, keepdims=keepdims, todense=True)
        if not issparse(ret):
            xps = get_sparse_module(self.spmatrix)
            ret = SparseMatrix(xps.csr_matrix(ret))
            return ret
        return ret

    def mean(self, axis=None, dtype=None, keepdims=None):
        return self._reduction('mean', axis=axis, dtype=dtype, keepdims=keepdims)

    def nansum(self, axis=None, dtype=None, keepdims=None):
        return self._reduction('nansum', axis=axis, dtype=dtype, keepdims=keepdims, todense=True)

    def nanprod(self, axis=None, dtype=None, keepdims=None):
        return self._reduction('nanprod', axis=axis, dtype=dtype, keepdims=keepdims, todense=True)

    def nanmax(self, axis=None, dtype=None, keepdims=None):
        ret = self._reduction('nanmax', axis=axis, dtype=dtype, keepdims=keepdims, todense=True)
        if not issparse(ret):
            xps = get_sparse_module(self.spmatrix)
            ret = SparseMatrix(xps.csr_matrix(ret))
            return ret
        return ret

    def nanmin(self, axis=None, dtype=None, keepdims=None):
        ret = self._reduction('nanmin', axis=axis, dtype=dtype, keepdims=keepdims, todense=True)
        if not issparse(ret):
            xps = get_sparse_module(self.spmatrix)
            ret = SparseMatrix(xps.csr_matrix(ret))
            return ret
        return ret

    def nanmean(self, axis=None, dtype=None, keepdims=None):
        return self._reduction('nanmean', axis=axis, dtype=dtype, keepdims=keepdims, todense=True)

    def argmax(self, axis=None, dtype=None, keepdims=None):
        return self._reduction('argmax', axis=axis, dtype=dtype, keepdims=keepdims)

    def nanargmax(self, axis=None, dtype=None, keepdims=None):
        return self._reduction('nanargmax', axis=axis, dtype=dtype, keepdims=keepdims, todense=True)

    def argmin(self, axis=None, dtype=None, keepdims=None):
        return self._reduction('argmin', axis=axis, dtype=dtype, keepdims=keepdims)

    def nanargmin(self, axis=None, dtype=None, keepdims=None):
        return self._reduction('nanargmin', axis=axis, dtype=dtype, keepdims=keepdims, todense=True)

    def var(self, axis=None, dtype=None, ddof=0, keepdims=None):
        return self._reduction('var', axis=axis, dtype=dtype, ddof=ddof, keepdims=keepdims, todense=True)

    def cumsum(self, axis=None, dtype=None):
        return self.spmatrix.toarray().cumsum(axis=axis)

    def cumprod(self, axis=None, dtype=None):
        return self.spmatrix.toarray().cumprod(axis=axis)

    def nancumsum(self, axis=None, dtype=None):
        xp = get_array_module(self.spmatrix)
        return xp.nancumsum(self.spmatrix.toarray(), axis=axis)

    def nancumprod(self, axis=None, dtype=None):
        xp = get_array_module(self.spmatrix)
        return xp.nancumprod(self.spmatrix.toarray(), axis=axis)

    def count_nonzero(self, axis=None, dtype=None, keepdims=None):
        if axis is None:
            return get_array_module(self.spmatrix).array([self.spmatrix.count_nonzero()])[0]
        else:
            return get_array_module(self.spmatrix).count_nonzero(self.spmatrix.toarray(), axis=axis)

    def __getitem__(self, item):
        if isinstance(item, SparseMatrix):
            item = item.spmatrix
        if isinstance(item, list):
            item = tuple(item)

        if not all(isinstance(i, slice) for i in item):
            raise NotImplementedError('sparse matrix only support slice for indexing')

        x = self.spmatrix[item]
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def __setitem__(self, key, value):
        if is_cupy(self.spmatrix):
            return NotImplemented
        else:
            x = self.spmatrix.tolil()
            x[key] = value
            x = x.tocsr()
        self.spmatrix = x

    def _maximum_minimum(self, other, method_name):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        if is_cupy(self.spmatrix):
            # TODO(jisheng): cupy does not implement sparse maximum and minimum
            return NotImplemented

        xps = get_sparse_module(self.spmatrix)
        xp = get_array_module(self.spmatrix)
        has_nan = xps.csr_matrix(
            (xp.isnan(self.spmatrix.data), self.spmatrix.indices, self.spmatrix.indptr),
            self.spmatrix.shape)
        if issparse(other):
            has_nan += xps.csr_matrix(
                (xp.isnan(other.data), other.indices, other.indptr), other.shape)

        x = getattr(self.spmatrix, method_name)(other)
        if has_nan.sum() > 0:
            x = x + (has_nan * np.nan)

        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def maximum(self, other):
        return self._maximum_minimum(other, 'maximum')

    def minimum(self, other):
        return self._maximum_minimum(other, 'minimum')

    def fmax(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        x = self.spmatrix.maximum(other)
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def fmin(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        x = self.spmatrix.minimum(other)
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def isinf(self):
        xp = get_array_module(self.spmatrix)
        data = xp.isinf(self.spmatrix.data)
        x = get_sparse_module(self.spmatrix).csr_matrix(
            (data, self.spmatrix.indices, self.spmatrix.indptr), self.spmatrix.shape
        )
        return SparseMatrix(x)

    def isnan(self):
        xp = get_array_module(self.spmatrix)
        data = xp.isnan(self.spmatrix.data)
        x = get_sparse_module(self.spmatrix).csr_matrix(
            (data, self.spmatrix.indices, self.spmatrix.indptr), self.spmatrix.shape
        )
        return SparseMatrix(x)

    def signbit(self):
        xp = get_array_module(self.spmatrix)
        data = xp.signbit(self.spmatrix.data)
        x = get_sparse_module(self.spmatrix).csr_matrix(
            (data, self.spmatrix.indices, self.spmatrix.indptr), self.spmatrix.shape
        )
        return SparseMatrix(x)

    def floor(self):
        return SparseMatrix(self.spmatrix.floor())

    def ceil(self):
        return SparseMatrix(self.spmatrix.ceil())

    def trunc(self):
        return SparseMatrix(self.spmatrix.trunc())

    def degrees(self):
        xp = get_array_module(self.spmatrix)
        data = xp.degrees(self.spmatrix.data)
        x = get_sparse_module(self.spmatrix).csr_matrix(
            (data, self.spmatrix.indices, self.spmatrix.indptr), self.spmatrix.shape
        )
        return SparseMatrix(x)

    def radians(self):
        xp = get_array_module(self.spmatrix)
        data = xp.radians(self.spmatrix.data)
        x = get_sparse_module(self.spmatrix).csr_matrix(
            (data, self.spmatrix.indices, self.spmatrix.indptr), self.spmatrix.shape
        )
        return SparseMatrix(x)

    def clip(self, a_min, a_max):
        try:
            a_min = naked(a_min)
        except TypeError:
            return NotImplemented

        try:
            a_max = naked(a_max)
        except TypeError:
            return NotImplemented

        x = self.spmatrix.maximum(a_min)
        if issparse(x):
            x = x.minimum(a_max)
        elif issparse(a_max):
            x = a_max.minimum(x)
        else:
            xp = get_array_module(x)
            x = xp.minimum(x, a_max)
        if issparse(x):
            return SparseMatrix(x)
        return get_array_module(x).asarray(x)

    def iscomplex(self):
        xp = get_array_module(self.spmatrix)
        data = xp.iscomplex(self.spmatrix.data)
        x = get_sparse_module(self.spmatrix).csr_matrix(
            (data, self.spmatrix.indices, self.spmatrix.indptr), self.spmatrix.shape
        )
        return SparseMatrix(x)

    def fix(self):
        xp = get_array_module(self.spmatrix)
        data = xp.fix(self.spmatrix.data)
        x = get_sparse_module(self.spmatrix).csr_matrix(
            (data, self.spmatrix.indices, self.spmatrix.indptr), self.spmatrix.shape
        )
        return SparseMatrix(x)

    def i0(self):
        xp = get_array_module(self.spmatrix)
        data = xp.i0(self.spmatrix.data).reshape(self.spmatrix.data.shape)
        x = get_sparse_module(self.spmatrix).csr_matrix(
            (data, self.spmatrix.indices, self.spmatrix.indptr), self.spmatrix.shape
        )
        return SparseMatrix(x)

    def nan_to_num(self):
        xp = get_array_module(self.spmatrix)
        data = xp.nan_to_num(self.spmatrix.data)
        x = get_sparse_module(self.spmatrix).csr_matrix(
            (data, self.spmatrix.indices, self.spmatrix.indptr), self.spmatrix.shape
        )
        return SparseMatrix(x)

    def copysign(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        if issparse(other):
            other = other.toarray()

        xp = get_array_module(self.spmatrix)
        return xp.copysign(self.spmatrix.toarray(), other)

    def nextafter(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        ret_sparse = False
        if issparse(other):
            ret_sparse = True
            other = other.toarray()

        xp = get_array_module(self.spmatrix)
        xps = get_sparse_module(self.spmatrix)

        x = xp.nextafter(self.spmatrix.toarray(), other)
        if ret_sparse:
            return SparseMatrix(xps.csr_matrix(x))
        return x

    def spacing(self):
        return get_array_module(self.spmatrix).spacing(self.spmatrix.toarray())

    def ldexp(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        if issparse(other):
            other = other.toarray()

        return SparseMatrix(self.spmatrix.multiply(2 ** other))

    def frexp(self, **kw):
        xp = get_array_module(self.spmatrix)
        xps = get_sparse_module(self.spmatrix)
        x, y = xp.frexp(self.spmatrix.toarray(), **kw)
        return SparseMatrix(xps.csr_matrix(x)), SparseMatrix(xps.csr_matrix(y))

    def modf(self, **kw):
        xp = get_array_module(self.spmatrix)
        xps = get_sparse_module(self.spmatrix)
        x, y = xp.modf(self.spmatrix.toarray(), **kw)
        return SparseMatrix(xps.csr_matrix(x)), SparseMatrix(xps.csr_matrix(y))

    def sinc(self):
        xp = get_array_module(self.spmatrix)
        return xp.sinc(self.spmatrix.toarray())

    def isfinite(self):
        xp = get_array_module(self.spmatrix)
        return xp.isfinite(self.spmatrix.toarray())

    def isreal(self):
        xp = get_array_module(self.spmatrix)
        return xp.isreal(self.spmatrix.toarray())

    def digitize(self, bins, right=False):
        xp = get_array_module(self.spmatrix)
        data = xp.digitize(self.spmatrix.data, bins, right)
        x = get_sparse_module(self.spmatrix).csr_matrix(
            (data, self.spmatrix.indices, self.spmatrix.indptr), self.spmatrix.shape
        )
        return SparseMatrix(x)

    def repeat(self, repeats, axis=None):
        if axis is None:
            raise NotImplementedError

        xp = get_array_module(self.spmatrix)
        xps = get_sparse_module(self.spmatrix)
        x = xps.csr_matrix(xp.repeat(self.spmatrix.toarray(), repeats, axis=axis))
        return SparseMatrix(x)



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import base64
import json
import pickle
import uuid
import logging

from tornado import gen, concurrent, web, ioloop

from .server import register_api_handler
from ..compat import six, futures
from ..lib.tblib import pickling_support
from ..actors import new_client
from .server import MarsWebAPI

pickling_support.install()
_actor_client = new_client()

logger = logging.getLogger(__name__)


def _patch_futures():
    _FUTURES = concurrent.FUTURES + (futures.Future, )

    def _is_future(x):
        return isinstance(x, _FUTURES)

    gen.is_future = _is_future
    ioloop.is_future = _is_future


if six.PY2:
    _patch_futures()


class ApiRequestHandler(web.RequestHandler):
    def initialize(self, scheduler_ip):
        self._scheduler = scheduler_ip
        self.web_api = MarsWebAPI(scheduler_ip)


class ApiEntryHandler(ApiRequestHandler):
    def get(self):
        self.write(dict(msg='Mars API Entry'))


class SessionsApiHandler(ApiRequestHandler):
    def post(self):
        args = {k: self.get_argument(k) for k in self.request.arguments}
        session_id = str(uuid.uuid1())
        self.web_api.create_session(session_id, **args)
        self.write(json.dumps(dict(session_id=session_id)))


class SessionApiHandler(ApiRequestHandler):
    def delete(self, session_id):
        self.web_api.delete_session(session_id)


class GraphsApiHandler(ApiRequestHandler):
    def post(self, session_id):
        try:
            graph = self.get_argument('graph')
            target = self.get_argument('target').split(',')
        except web.MissingArgumentError as ex:
            self.write(json.dumps(dict(msg=str(ex))))
            raise web.HTTPError(400, 'Argument missing')

        try:
            graph_key = str(uuid.uuid4())
            self.web_api.submit_graph(session_id, graph, graph_key, target)
            self.write(json.dumps(dict(graph_key=graph_key)))
        except:  # noqa: E722
            pickled_exc = pickle.dumps(sys.exc_info())
            self.write(json.dumps(dict(
                exc_info=base64.b64encode(pickled_exc),
            )))
            raise web.HTTPError(500, 'Internal server error')


class GraphApiHandler(ApiRequestHandler):
    @gen.coroutine
    def get(self, session_id, graph_key):
        from ..scheduler.utils import GraphState

        state = self.web_api.get_graph_state(session_id, graph_key)
        if state == GraphState.RUNNING:
            self.write(json.dumps(dict(state='running')))
        elif state == GraphState.SUCCEEDED:
            self.write(json.dumps(dict(state='success')))
        elif state == GraphState.FAILED:
            self.write(json.dumps(dict(state='failed')))
        elif state == GraphState.CANCELLED:
            self.write(json.dumps(dict(state='cancelled')))
        elif state == GraphState.CANCELLING:
            self.write(json.dumps(dict(state='cancelling')))
        elif state == GraphState.PREPARING:
            self.write(json.dumps(dict(state='preparing')))

    def delete(self, session_id, graph_key):
        self.web_api.stop_graph(session_id, graph_key)


class GraphDataHandler(ApiRequestHandler):
    @gen.coroutine
    def get(self, session_id, graph_key, tensor_key):
        executor = futures.ThreadPoolExecutor(1)

        def _fetch_fun():
            web_api = MarsWebAPI(self._scheduler)
            return web_api.fetch_data(session_id, graph_key, tensor_key)

        data = yield executor.submit(_fetch_fun)
        self.write(data)

    def delete(self, session_id, graph_key, tensor_key):
        self.web_api.delete_data(session_id, graph_key, tensor_key)


class WorkersApiHandler(ApiRequestHandler):
    def get(self):
        workers_num = self.web_api.count_workers()
        self.write(json.dumps(workers_num))


register_api_handler('/api', ApiEntryHandler)
register_api_handler('/api/session', SessionsApiHandler)
register_api_handler('/api/worker', WorkersApiHandler)
register_api_handler('/api/session/(?P<session_id>[^/]+)', SessionApiHandler)
register_api_handler('/api/session/(?P<session_id>[^/]+)/graph', GraphsApiHandler)
register_api_handler('/api/session/(?P<session_id>[^/]+)/graph/(?P<graph_key>[^/]+)', GraphApiHandler)
register_api_handler('/api/session/(?P<session_id>[^/]+)/graph/(?P<graph_key>[^/]+)/data/(?P<tensor_key>[^/]+)',
                     GraphDataHandler)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from ..utils import inject_dtype
from .core import TensorCompare, TensorCompareConstant


class TensorNotEqual(operands.NotEqual, TensorCompare):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorNotEqual, self).__init__(_casting=casting, _err=err,
                                             _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def constant_cls(cls):
        return TensorNeConstant


class TensorNeConstant(operands.NeConstant, TensorCompareConstant):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        super(TensorNeConstant, self).__init__(_casting=casting, _err=err,
                                               _dtype=dtype, _sparse=sparse, **kw)


@inject_dtype(np.bool_)
def not_equal(x1, x2, out=None, where=None, **kwargs):
    """
    Return (x1 != x2) element-wise.

    Parameters
    ----------
    x1, x2 : array_like
      Input tensors.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    not_equal : tensor bool, scalar bool
      For each element in `x1, x2`, return True if `x1` is not equal
      to `x2` and False otherwise.


    See Also
    --------
    equal, greater, greater_equal, less, less_equal

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.not_equal([1.,2.], [1., 3.]).execute()
    array([False,  True])
    >>> mt.not_equal([1, 2], [[1, 3],[1, 4]]).execute()
    array([[False,  True],
           [False,  True]])
    """
    op = TensorNotEqual(**kwargs)
    return op(x1, x2, out=out, where=where)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .core import TensorRandomOperandMixin


class TensorStandardNormal(operands.StandardNormal, TensorRandomOperandMixin):
    __slots__ = '_size',

    def __init__(self, size=None, state=None, dtype=None, gpu=None, **kw):
        dtype = np.dtype(dtype) if dtype is not None else dtype
        super(TensorStandardNormal, self).__init__(_size=size, _state=state, _dtype=dtype,
                                                   _gpu=gpu, **kw)

    def __call__(self, chunk_size=None):
        return self.new_tensor(None, None, raw_chunk_size=chunk_size)


def standard_normal(random_state, size=None, chunk_size=None, gpu=None, **kw):
    """
    Draw samples from a standard Normal distribution (mean=0, stdev=1).

    Parameters
    ----------
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  Default is None, in which case a
        single value is returned.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    -------
    out : float or Tensor
        Drawn samples.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> s = mt.random.standard_normal(8000)
    >>> s.execute()
    array([ 0.6888893 ,  0.78096262, -0.89086505, ...,  0.49876311, #random
           -0.38672696, -0.4685006 ])                               #random
    >>> s.shape
    (8000,)
    >>> s = mt.random.standard_normal(size=(3, 4, 2))
    >>> s.shape
    (3, 4, 2)
    """
    if 'dtype' not in kw:
        kw['dtype'] = np.random.RandomState().standard_normal(size=(0,)).dtype
    size = random_state._handle_size(size)
    op = TensorStandardNormal(size=size, state=random_state._state, gpu=gpu, **kw)
    return op(chunk_size=chunk_size)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from ..datasource import tensor as astensor
from .core import TensorReduction


class TensorMoment(operands.Moment, TensorReduction):
    def __init__(self, axis=None, dtype=None, keepdims=None, moment=None, ddof=None, combine_size=None, **kw):
        if moment is not None:
            kw['_moment'] = moment
        super(TensorMoment, self).__init__(_axis=axis, _dtype=dtype, _keepdims=keepdims,
                                           _ddof=ddof, _combine_size=combine_size, **kw)


class TensorMomentChunk(operands.MomentChunk, TensorReduction):
    def __init__(self, axis=None, dtype=None, keepdims=None, moment=None, combine_size=None, **kw):
        if moment is not None:
            kw['_moment'] = moment
        super(TensorMomentChunk, self).__init__(_axis=axis, _dtype=dtype, _keepdims=keepdims,
                                                _combine_size=combine_size, **kw)


class TensorMomentCombine(operands.MomentCombine, TensorReduction):
    def __init__(self, axis=None, dtype=None, keepdims=None, moment=None, combine_size=None, **kw):
        if moment is not None:
            kw['_moment'] = moment
        super(TensorMomentCombine, self).__init__(_axis=axis, _dtype=dtype, _keepdims=keepdims,
                                                  _combine_size=combine_size, **kw)


class TensorVar(operands.Var, TensorReduction):
    def __init__(self, axis=None, dtype=None, keepdims=None, ddof=0, combine_size=None, **kw):
        super(TensorVar, self).__init__(_axis=axis, _dtype=dtype, _keepdims=keepdims, _ddof=ddof,
                                        _combine_size=combine_size, **kw)

    @staticmethod
    def _get_op_types():
        return TensorMomentChunk, TensorMoment, TensorMomentCombine

    def _get_op_kw(self):
        kw = dict()
        kw['ddof'] = self.ddof
        return kw


def var(a, axis=None, dtype=None, out=None, ddof=0, keepdims=None, combine_size=None):
    """
    Compute the variance along the specified axis.

    Returns the variance of the tensor elements, a measure of the spread of a
    distribution.  The variance is computed for the flattened tensor by
    default, otherwise over the specified axis.

    Parameters
    ----------
    a : array_like
        Tensor containing numbers whose variance is desired.  If `a` is not a
        tensor, a conversion is attempted.
    axis : None or int or tuple of ints, optional
        Axis or axes along which the variance is computed.  The default is to
        compute the variance of the flattened array.

        If this is a tuple of ints, a variance is performed over multiple axes,
        instead of a single axis or all the axes as before.
    dtype : data-type, optional
        Type to use in computing the variance.  For arrays of integer type
        the default is `float32`; for tensors of float types it is the same as
        the tensor type.
    out : Tensor, optional
        Alternate output array in which to place the result.  It must have
        the same shape as the expected output, but the type is cast if
        necessary.
    ddof : int, optional
        "Delta Degrees of Freedom": the divisor used in the calculation is
        ``N - ddof``, where ``N`` represents the number of elements. By
        default `ddof` is zero.
    keepdims : bool, optional
        If this is set to True, the axes which are reduced are left
        in the result as dimensions with size one. With this option,
        the result will broadcast correctly against the input tensor.

        If the default value is passed, then `keepdims` will not be
        passed through to the `var` method of sub-classes of
        `Tensor`, however any non-default value will be.  If the
        sub-classes `sum` method does not implement `keepdims` any
        exceptions will be raised.
    combine_size: int, optional
        The number of chunks to combine.

    Returns
    -------
    variance : Tensor, see dtype parameter above
        If ``out=None``, returns a new tensor containing the variance;
        otherwise, a reference to the output tensor is returned.

    See Also
    --------
    std , mean, nanmean, nanstd, nanvar

    Notes
    -----
    The variance is the average of the squared deviations from the mean,
    i.e.,  ``var = mean(abs(x - x.mean())**2)``.

    The mean is normally calculated as ``x.sum() / N``, where ``N = len(x)``.
    If, however, `ddof` is specified, the divisor ``N - ddof`` is used
    instead.  In standard statistical practice, ``ddof=1`` provides an
    unbiased estimator of the variance of a hypothetical infinite population.
    ``ddof=0`` provides a maximum likelihood estimate of the variance for
    normally distributed variables.

    Note that for complex numbers, the absolute value is taken before
    squaring, so that the result is always real and nonnegative.

    For floating-point input, the variance is computed using the same
    precision the input has.  Depending on the input data, this can cause
    the results to be inaccurate, especially for `float32` (see example
    below).  Specifying a higher-accuracy accumulator using the ``dtype``
    keyword can alleviate this issue.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = mt.array([[1, 2], [3, 4]])
    >>> mt.var(a).execute()
    1.25
    >>> mt.var(a, axis=0).execute()
    array([ 1.,  1.])
    >>> mt.var(a, axis=1).execute()
    array([ 0.25,  0.25])

    In single precision, var() can be inaccurate:

    >>> a = mt.zeros((2, 512*512), dtype=mt.float32)
    >>> a[0, :] = 1.0
    >>> a[1, :] = 0.1
    >>> mt.var(a).execute()
    0.20250003

    Computing the variance in float64 is more accurate:

    >>> mt.var(a, dtype=mt.float64).execute()
    0.20249999932944759
    >>> ((1-0.55)**2 + (0.1-0.55)**2)/2
    0.2025

    """
    a = astensor(a)
    if dtype is None:
        dtype = np.var(np.ones((1,), dtype=a.dtype)).dtype
    op = TensorVar(axis=axis, dtype=dtype, keepdims=keepdims, ddof=ddof, combine_size=combine_size)
    return op(a, out=out)



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from .actors import new_client
from .cluster_info import ClusterInfoActor
from .node_info import NodeInfoActor
from .scheduler import SessionActor, GraphActor, GraphMetaActor, ResourceActor
from .scheduler.session import SessionManagerActor
from .scheduler.graph import ResultReceiverActor

logger = logging.getLogger(__name__)


class MarsAPI(object):
    def __init__(self, scheduler_ip):
        self.actor_client = new_client()
        self.cluster_info = self.actor_client.actor_ref(
            ClusterInfoActor.default_name(), address=scheduler_ip)
        self.session_manager = self.get_actor_ref(SessionManagerActor.default_name())

    def get_actor_ref(self, uid):
        actor_address = self.cluster_info.get_scheduler(uid)
        return self.actor_client.actor_ref(uid, address=actor_address)

    def get_schedulers_info(self):
        schedulers = self.cluster_info.get_schedulers()
        infos = dict()
        for scheduler in schedulers:
            info_ref = self.actor_client.actor_ref(NodeInfoActor.default_name(),
                                                   address=scheduler)
            infos[scheduler] = info_ref.get_info()
        return infos

    def count_workers(self):
        try:
            uid = ResourceActor.default_name()
            return self.get_actor_ref(uid).get_worker_count()
        except KeyError:
            return 0

    def create_session(self, session_id, **kw):
        self.session_manager.create_session(session_id, **kw)

    def delete_session(self, session_id):
        self.session_manager.delete_session(session_id)

    def submit_graph(self, session_id, serialized_graph, graph_key, target):
        session_uid = SessionActor.gen_name(session_id)
        session_ref = self.get_actor_ref(session_uid)
        session_ref.submit_tensor_graph(serialized_graph, graph_key, target, _tell=True)

    def delete_graph(self, session_id, graph_key):
        graph_uid = GraphActor.gen_name(session_id, graph_key)
        graph_ref = self.get_actor_ref(graph_uid)
        graph_ref.destroy()

    def stop_graph(self, session_id, graph_key):
        from .scheduler import GraphState
        graph_meta_uid = GraphMetaActor.gen_name(session_id, graph_key)
        self.get_actor_ref(graph_meta_uid).set_state(GraphState.CANCELLING)

        graph_uid = GraphActor.gen_name(session_id, graph_key)
        graph_ref = self.get_actor_ref(graph_uid)
        graph_ref.stop_graph()

    def get_graph_state(self, session_id, graph_key):
        from .scheduler import GraphState

        graph_meta_uid = GraphMetaActor.gen_name(session_id, graph_key)
        graph_meta_ref = self.get_actor_ref(graph_meta_uid)
        if self.actor_client.has_actor(graph_meta_ref):
            state_obj = graph_meta_ref.get_state()
            state = state_obj.value if state_obj else 'preparing'
        else:
            state = 'preparing'
        state = GraphState(state.lower())
        return state

    def fetch_data(self, session_id, graph_key, tensor_key):
        graph_uid = GraphActor.gen_name(session_id, graph_key)
        graph_address = self.cluster_info.get_scheduler(graph_uid)
        result_ref = self.actor_client.create_actor(ResultReceiverActor, address=graph_address)
        return result_ref.fetch_tensor(session_id, graph_key, tensor_key)

    def delete_data(self, session_id, graph_key, tensor_key):
        graph_uid = GraphActor.gen_name(session_id, graph_key)
        graph_ref = self.get_actor_ref(graph_uid)
        graph_ref.free_tensor_data(tensor_key, _tell=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
import json
import time

from ...api import MarsAPI
from ...compat import TimeoutError
from ...graph import DirectedGraph
from ...scheduler.graph import GraphState
from ...serialize import dataserializer


class LocalClusterSession(object):
    def __init__(self, endpoint):
        self._session_id = uuid.uuid4()
        self._endpoint = endpoint
        self._tensor_to_graph = dict()
        self._api = MarsAPI(self._endpoint)

        # create session on the cluster side
        self._api.create_session(self._session_id)

    @property
    def endpoint(self):
        return self._endpoint

    @endpoint.setter
    def endpoint(self, endpoint):
        self._endpoint = endpoint
        self._api = MarsAPI(self._endpoint)

    def run(self, *tensors, **kw):
        timeout = kw.pop('timeout', -1)
        if kw:
            raise TypeError('run got unexpected key arguments {0}'.format(', '.join(kw.keys())))

        graph = DirectedGraph()
        targets = [t.key for t in tensors]
        graph_key = uuid.uuid4()
        for t in tensors:
            t.build_graph(graph, tiled=False)
            self._tensor_to_graph[t.key] = graph_key

        # submit graph to local cluster
        self._api.submit_graph(self._session_id, json.dumps(graph.to_json()),
                               graph_key, targets)

        exec_start_time = time.time()
        while timeout <= 0 or time.time() - exec_start_time <= timeout:
            time.sleep(0.1)

            graph_state = self._api.get_graph_state(self._session_id, graph_key)
            if graph_state == GraphState.SUCCEEDED:
                break
            if graph_state == GraphState.FAILED:
                # TODO(qin): add traceback
                raise SystemError('Graph execution failed with unknown reason')

        if 0 < timeout < time.time() - exec_start_time:
            raise TimeoutError

        data_list = []
        for target in targets:
            resp = self._api.fetch_data(self._session_id, graph_key, target)
            data_list.append(dataserializer.loads(resp))

        return data_list

    def fetch(self, tensor):
        key = tensor.key
        graph_key = self._tensor_to_graph[key]
        resp = self._api.fetch_data(self._session_id, graph_key, key)
        return dataserializer.loads(resp)

    def decref(self, *keys):
        for k in keys:
            if k not in self._tensor_to_graph:
                continue
            graph_key = self._tensor_to_graph[k]
            self._api.delete_data(self._session_id, graph_key, k)

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self._api.delete_session(self._session_id)



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import copy
import logging
import time

from .assigner import AssignerActor
from .chunkmeta import ChunkMetaActor
from .graph import GraphActor
from .kvstore import KVStoreActor
from .resource import ResourceActor
from .utils import SchedulerActor, OperandState, GraphState, array_to_bytes
from ..compat import six
from ..config import options
from ..errors import *
from ..utils import log_unhandled

logger = logging.getLogger(__name__)


class OperandActor(SchedulerActor):
    """
    Actor handling the whole lifecycle of a particular operand instance
    """
    @staticmethod
    def gen_uid(session_id, op_key):
        return 's:operator$%s$%s' % (session_id, op_key)

    def __init__(self, session_id, graph_id, op_key, op_info, worker_endpoint=None,
                 is_terminal=False):
        super(OperandActor, self).__init__()
        op_info = copy.deepcopy(op_info)

        self._session_id = session_id
        self._graph_id = graph_id
        self._op_key = op_key
        self._op_path = '/sessions/%s/operands/%s' % (self._session_id, self._op_key)

        self._cluster_info_ref = None
        self._assigner_ref = None
        self._graph_ref = None
        self._resource_ref = None
        self._kv_store_ref = None
        self._chunk_meta_ref = None

        self._op_name = op_info['op_name']
        self._state = OperandState(op_info['state'].lower())
        self._last_state = self._state
        self._retries = op_info['retries']
        self._is_terminal = is_terminal
        self._io_meta = op_info['io_meta']
        self._pred_keys = self._io_meta['predecessors']
        self._succ_keys = self._io_meta['successors']
        self._input_chunks = self._io_meta['input_chunks']
        self._chunks = self._io_meta['chunks']
        self._info = op_info

        # worker the operand expected to be executed on
        self._expect_worker = op_info.get('target_worker')
        # worker actually assigned
        self._worker_endpoint = worker_endpoint

        # ref of ExecutionActor on worker
        self._execution_ref = None
        # set of finished predecessors, used to decide whether we should move the operand to ready
        self._finish_preds = set()
        # set of finished successors, used to detect whether we can do clean up
        self._finish_succs = set()

        self._input_worker_scores = dict()
        self._worker_scores = dict()

        # handlers of states. will be called when the state of the operand switches
        # from one to another
        self._state_handlers = {
            OperandState.READY: self._handle_ready,
            OperandState.RUNNING: self._handle_running,
            OperandState.FINISHED: self._handle_finished,
            OperandState.FREED: self._handle_freed,
            OperandState.FATAL: self._handle_fatal,
            OperandState.CANCELLING: self._handle_cancelling,
            OperandState.CANCELLED: self._handle_cancelled,
        }

    def post_create(self):
        self.set_cluster_info_ref()
        self._assigner_ref = self.get_promise_ref(AssignerActor.gen_name(self._session_id))
        self._chunk_meta_ref = self.ctx.actor_ref(ChunkMetaActor.default_name())
        self._graph_ref = self.get_actor_ref(GraphActor.gen_name(self._session_id, self._graph_id))
        self._resource_ref = self.get_actor_ref(ResourceActor.default_name())

        self._kv_store_ref = self.ctx.actor_ref(KVStoreActor.default_name())
        if not self.ctx.has_actor(self._kv_store_ref):
            self._kv_store_ref = None

    def add_finished_predecessor(self, op_key):
        self._finish_preds.add(op_key)
        if all(k in self._finish_preds for k in self._pred_keys):
            if self.state in (OperandState.CANCELLED, OperandState.CANCELLING):
                return True
            # all predecessors done, the operand can be executed now
            self.state = OperandState.READY
            self.start_operand()
            return True
        self.update_demand_depths(self._info.get('optimize', {}).get('depth', 0))
        return False

    def add_finished_successor(self, op_key):
        self._finish_succs.add(op_key)
        if not self._is_terminal and all(k in self._finish_succs for k in self._succ_keys):
            # non-terminal operand with all successors done, the data can be freed
            self.ref().free_data(_tell=True)

    def update_demand_depths(self, depth):
        """
        Update the depth of operand demanding data, or demanding the descendant of data
        produced by the current operand
        :param depth: depth to update
        """
        demand_depths = list(self._info.get('optimize', {}).get('demand_depths', ()))
        if not demand_depths:
            demand_depths = [depth]
        else:
            idx = 0
            for idx, v in enumerate(demand_depths):
                if v <= depth:
                    break
            if demand_depths[idx] == depth:
                return
            elif demand_depths[idx] > depth:
                demand_depths.append(depth)
            else:
                demand_depths.insert(idx, depth)
        try:
            optimize_data = self._info['optimize']
        except KeyError:
            optimize_data = self._info['optimize'] = dict()
        optimize_data['demand_depths'] = tuple(demand_depths)
        if self._kv_store_ref is not None:
            self._kv_store_ref.write(
                '%s/optimize/demand_depths' % self._op_path,
                base64.b64encode(array_to_bytes('I', demand_depths)), _tell=True, _wait=False)
        if self.state == OperandState.READY:
            # if the operand is already submitted to AssignerActor, we need to update the priority
            self._assigner_ref.update_priority(self._op_key, self._info['optimize'], _tell=True, _wait=False)
        else:
            # send update command to predecessors
            for in_key in self._pred_keys:
                self._get_operand_actor(in_key).update_demand_depths(depth, _tell=True, _wait=False)

    def propose_descendant_workers(self, input_key, worker_scores, depth=1):
        """
        Calculate likelihood of the operand being sent to workers
        :param input_key: input key that carries the scores
        :param worker_scores: score of every worker on input key
        :param depth: maximal propagate depth
        """
        if self._worker_endpoint:
            # worker already assigned, there should be no other possibilities
            self._worker_scores = {self._worker_endpoint: 1.0}
        elif self._expect_worker:
            # worker already propsed, there should be no other possibilities
            self._worker_scores = {self._expect_worker: 1.0}
        else:
            # aggregate the score from input to the score of current operand
            old_scores = self._input_worker_scores.get(input_key, {})
            self._input_worker_scores[input_key] = worker_scores
            all_keys = set(old_scores.keys()) | set(worker_scores)
            for k in all_keys:
                delta = (worker_scores.get(k, 0) - old_scores.get(k, 0)) * 1.0 / len(self._pred_keys)
                self._worker_scores[k] = self._worker_scores.get(k, 0) + delta
                if self._worker_scores[k] < 1e-6:
                    del self._worker_scores[k]

        if depth:
            # push down to successors
            for succ_key in self._succ_keys:
                self._get_operand_actor(succ_key).propose_descendant_workers(
                    self._op_key, self._worker_scores, depth=depth - 1, _tell=True)
        # pick the worker with largest likelihood
        max_score = 0
        max_worker = None
        for k, v in self._worker_scores.items():
            if v > max_score:
                max_score = v
                max_worker = k
        if max_score > 0.5:
            logger.debug('Operand %s(%s) now owning a dominant worker %s. scores=%r',
                         self._op_key, self._op_name, max_worker, self._worker_scores)
            return self._input_chunks, max_worker

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        self._last_state = self._state
        if value != self._last_state:
            logger.debug('Operand %s(%s) state from %s to %s.', self._op_key, self._op_name,
                         self._last_state, value)
        self._state = value
        self._info['state'] = value.name
        futures = [
            self._graph_ref.set_operand_state(self._op_key, value.value, _tell=True, _wait=False),
        ]
        if self._kv_store_ref is not None:
            futures.append(self._kv_store_ref.write('%s/state' % self._op_path, value.name, _tell=True, _wait=False))
        [f.result() for f in futures]

    @property
    def retries(self):
        return self._retries

    @retries.setter
    def retries(self, value):
        futures = []
        self._retries = value
        self._info['retries'] = value

        if self._kv_store_ref is not None:
            futures.append(self._kv_store_ref.write('%s/retries' % self._op_path, str(value), _tell=True, _wait=False))

        retry_timestamp = time.time()
        self._info['retry_timestamp'] = retry_timestamp
        if self._kv_store_ref is not None:
            futures.append(self._kv_store_ref.write('%s/retry_timestamp' % self._op_path, str(value),
                                                    _tell=True, _wait=False))
        [f.result() for f in futures]

    def get_op_info(self):
        info = dict()
        info['name'] = self._op_name
        info['state'] = self.state
        return info

    def start_operand(self):
        """
        Start handling operand given self.state
        """
        self._state_handlers[self.state]()

    def stop_operand(self):
        """
        Stop operand by starting CANCELLING procedure
        """
        if self.state == OperandState.CANCELLING or self.state == OperandState.CANCELLED:
            return
        self.state = OperandState.CANCELLING
        self.start_operand()

    def _free_worker_data(self, ep, chunk_key):
        """
        Free data on single worker
        :param ep: worker endpoint
        :param chunk_key: chunk key
        """
        from ..worker.chunkholder import ChunkHolderActor

        worker_cache_ref = self.ctx.actor_ref(ep, ChunkHolderActor.default_name())
        return worker_cache_ref.unregister_chunk(self._session_id, chunk_key,
                                                 _tell=True, _wait=False)

    def free_data(self, state=OperandState.FREED):
        """
        Free output data of current operand
        :param state: target state
        """
        if self.state == OperandState.FREED:
            return
        endpoint_lists = self._chunk_meta_ref.batch_get_workers(self._session_id, self._chunks)
        futures = []
        for chunk_key, endpoints in zip(self._chunks, endpoint_lists):
            if endpoints is None:
                continue
            for ep in endpoints:
                futures.append(self._free_worker_data(ep, chunk_key))
        futures.append(self._chunk_meta_ref.batch_delete_meta(
            self._session_id, self._chunks, _tell=True, _wait=False))
        [f.result() for f in futures]
        self.state = state
        self.start_operand()

    def propagate_state(self, state):
        """
        Change state of current operand by other operands
        :param state: target state
        """
        if self.state == OperandState.CANCELLING or self.state == OperandState.CANCELLED:
            return
        if self.state != state:
            self.state = state
            self.start_operand()

    def _get_raw_execution_ref(self):
        """
        Get raw ref of ExecutionActor on assigned worker. This method can be patched on debug
        """
        from ..worker.dispatcher import DispatchActor

        dispatch_ref = self.promise_ref(DispatchActor.default_name(), address=self._worker_endpoint)
        exec_uid = dispatch_ref.get_hash_slot('execution', self._op_key)
        return self.ctx.actor_ref(exec_uid, address=self._worker_endpoint)

    def _get_execution_ref(self):
        """
        Get ref of ExecutionActor on assigned worker
        """
        if self._execution_ref is None:
            self._execution_ref = self.promise_ref(self._get_raw_execution_ref())
        return self._execution_ref

    def _get_operand_actor(self, key):
        """
        Get ref of OperandActor by operand key
        """
        op_uid = self.gen_uid(self._session_id, key)
        return self.ctx.actor_ref(op_uid, address=self.get_scheduler(op_uid))

    @log_unhandled
    def _handle_ready(self):
        @log_unhandled
        def _submit_job(worker):
            # worker assigned, submit job
            if self.state in (OperandState.CANCELLED, OperandState.CANCELLING):
                self.start_operand()
                return

            self._worker_endpoint = worker

            target_predicts = dict()
            if options.scheduler.enable_active_push:
                # if active push enabled, we calculate the most possible target
                for succ_key in self._succ_keys:
                    succ_worker_predict = self._get_operand_actor(succ_key) \
                        .propose_descendant_workers(self._op_key, {worker: 1.0})
                    if not succ_worker_predict:
                        continue
                    keys, target = succ_worker_predict
                    if target == worker:
                        continue
                    for k in keys:
                        if k not in self._chunks:
                            continue
                        if k not in target_predicts:
                            target_predicts[k] = set()
                        target_predicts[k].add(target)
            if not target_predicts:
                target_predicts = None
            else:
                logger.debug('Receive active pushing list for operand %s: %r',
                             self._op_key, target_predicts)

            data_sizes = dict(zip(
                self._input_chunks,
                self._chunk_meta_ref.batch_get_chunk_size(self._session_id, self._input_chunks),
            ))

            # submit job
            self._execution_ref = self._get_execution_ref()
            serialized_exec_graph = self._graph_ref.get_executable_operand_dag(self._op_key)
            self._execution_ref.execute_graph(self._session_id, self._op_key, serialized_exec_graph,
                                              self._io_meta, data_sizes, send_targets=target_predicts,
                                              _promise=True)
            self.state = OperandState.RUNNING
            self.start_operand()

        logger.debug('Applying for resources for operand %r', self._info)
        # if under retry, give application a delay
        delay = options.scheduler.retry_delay if self.retries else 0
        # Send resource application. Submit job when worker assigned
        self._assigner_ref.apply_for_resource(
            self._session_id, self._op_key, self._info, _delay=delay, _promise=True) \
            .then(_submit_job)

    @log_unhandled
    def _handle_running(self):
        self._execution_ref = self._get_execution_ref()

        @log_unhandled
        def _acceptor(*_):
            # handling success of operand execution
            self._resource_ref.deallocate_resource(self._session_id, self._op_key, self._worker_endpoint)
            self.state = OperandState.FINISHED
            self.start_operand()

        @log_unhandled
        def _rejecter(*exc):
            # handling execption occurance of operand execution
            self._resource_ref.deallocate_resource(self._session_id, self._op_key, self._worker_endpoint)
            if exc and not isinstance(exc[0], type):
                raise TypeError('Unidentified rejection args: %r', exc)

            if self.state == OperandState.CANCELLING:
                logger.warning('Execution of operand %s interrupted.', self._op_key)
                self.free_data(OperandState.CANCELLED)
                return

            if exc and issubclass(exc[0], ResourceInsufficient):
                # resource insufficient: just set to READY and continue
                self.state = OperandState.READY
                self.ref().start_operand(_tell=True)
            elif exc and issubclass(exc[0], ExecutionInterrupted):
                # job cancelled: switch to cancelled
                self._resource_ref.deallocate_resource(self._session_id, self._op_key, self._worker_endpoint)
                logger.warning('Execution of operand %s interrupted.', self._op_key)
                self.free_data(OperandState.CANCELLED)
            else:
                self._resource_ref.deallocate_resource(self._session_id, self._op_key, self._worker_endpoint)
                try:
                    if exc:
                        six.reraise(*exc)
                    else:
                        raise SystemError('Worker throws rejection without details')
                except:
                    logger.exception('Attempt %d: Unexpected error occurred in executing operand %s in %s',
                                     self.retries + 1, self._op_key, self._worker_endpoint)
                # increase retry times
                self.retries += 1
                if self.retries >= options.scheduler.retry_num:
                    # no further trial
                    self.state = OperandState.FATAL
                else:
                    self.state = OperandState.READY
                self.start_operand()

        self._execution_ref.add_finish_callback(self._session_id, self._op_key, _promise=True) \
            .then(_acceptor, _rejecter)

    @log_unhandled
    def _handle_finished(self):
        self._resource_ref.deallocate_resource(self._session_id, self._op_key, self._worker_endpoint)
        if self._last_state == OperandState.CANCELLING:
            self.state = OperandState.CANCELLING
            self.start_operand()
            return
        new_ready_generated = False
        # update pred & succ finish records to trigger further actions
        # record if successors can be executed
        for out_key in self._succ_keys:
            new_ready_generated |= self._get_operand_actor(out_key).add_finished_predecessor(self._op_key)
        for in_key in self._pred_keys:
            self._get_operand_actor(in_key).add_finished_successor(self._op_key)
        # require more chunks to execute if the completion caused no successors to run
        if not new_ready_generated:
            self._assigner_ref.allocate_top_resources(_tell=True)
        if self._is_terminal:
            # update records in GraphActor to help decide if the whole graph finished execution
            self._graph_ref.mark_terminal_finished(self._op_key, _tell=True)

    @log_unhandled
    def _handle_fatal(self):
        if self._last_state == OperandState.FATAL:
            return
        if self._worker_endpoint:
            self._resource_ref.deallocate_resource(self._session_id, self._op_key, self._worker_endpoint)

        if self._is_terminal:
            # update records in GraphActor to help decide if the whole graph finished execution
            self._graph_ref.mark_terminal_finished(self._op_key, final_state=GraphState.FAILED, _tell=True)
        # set successors to FATAL
        for k in self._succ_keys:
            self._get_operand_actor(k).propagate_state(OperandState.FATAL, _tell=True)

    @log_unhandled
    def _handle_cancelling(self):
        if self._last_state == OperandState.CANCELLING:
            return
        if self._last_state == OperandState.CANCELLED:
            self.state = OperandState.CANCELLED
            return

        if self._last_state == OperandState.RUNNING:
            # send stop to worker
            self._execution_ref = self._get_execution_ref()
            self._execution_ref.stop_execution(self._op_key, _tell=True)
            return
        if self._last_state == OperandState.FINISHED:
            # delete data on cancelled
            self.ref().free_data(state=OperandState.CANCELLED, _tell=True)
            return
        if self._last_state == OperandState.READY:
            # stop worker application
            self._assigner_ref.remove_apply(self._op_key)
        self.state = OperandState.CANCELLED
        self.ref().start_operand(_tell=True)

    @log_unhandled
    def _handle_cancelled(self):
        if self._worker_endpoint:
            self._resource_ref.deallocate_resource(self._session_id, self._op_key, self._worker_endpoint)
        if self._is_terminal:
            self._graph_ref.mark_terminal_finished(self._op_key, final_state=GraphState.CANCELLED, _tell=True)
        for k in self._succ_keys:
            self._get_operand_actor(k).propagate_state(OperandState.CANCELLING, _tell=True)

    @log_unhandled
    def _handle_freed(self):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import operator

import numpy as np

from .array import as_same_device, device, is_sparse_module, get_array_module
from ..expressions import arithmetic
from ...compat import reduce, six

OP_TO_HANDLER = {
    arithmetic.TensorAdd: 'add',
    arithmetic.TensorSubtract: 'subtract',
    arithmetic.TensorMultiply: 'multiply',
    arithmetic.TensorDivide: 'divide',
    arithmetic.TensorTrueDiv: 'true_divide',
    arithmetic.TensorFloorDiv: 'floor_divide',
    arithmetic.TensorPower: 'power',
    arithmetic.TensorFloatPower: 'float_power',
    arithmetic.TensorMod: 'mod',
    arithmetic.TensorFMod: 'fmod',
    arithmetic.TensorLogAddExp: 'logaddexp',
    arithmetic.TensorLogAddExp2: 'logaddexp2',
    arithmetic.TensorNegative: 'negative',
    arithmetic.TensorPositive: operator.pos,
    arithmetic.TensorAbsolute: 'absolute',
    arithmetic.TensorAbs: 'abs',
    arithmetic.TensorFabs: 'fabs',
    arithmetic.TensorRint: 'rint',
    arithmetic.TensorSign: 'sign',
    arithmetic.TensorConj: 'conj',
    arithmetic.TensorExp: 'exp',
    arithmetic.TensorExp2: 'exp2',
    arithmetic.TensorLog: 'log',
    arithmetic.TensorLog2: 'log2',
    arithmetic.TensorLog10: 'log10',
    arithmetic.TensorExpm1: 'expm1',
    arithmetic.TensorLog1p: 'log1p',
    arithmetic.TensorSqrt: 'sqrt',
    arithmetic.TensorSquare: 'square',
    arithmetic.TensorCbrt: 'cbrt',
    arithmetic.TensorReciprocal: 'reciprocal',
    arithmetic.TensorAround: 'around',
    arithmetic.TensorIsFinite: 'isfinite',
    arithmetic.TensorIsInf: 'isinf',
    arithmetic.TensorIsNan: 'isnan',
    arithmetic.TensorSignbit: 'signbit',
    arithmetic.TensorCopysign: 'copysign',
    arithmetic.TensorNextafter: 'nextafter',
    arithmetic.TensorSpacing: 'spacing',
    arithmetic.TensorLdexp: 'ldexp',
    arithmetic.TensorFloor: 'floor',
    arithmetic.TensorCeil: 'ceil',
    arithmetic.TensorTrunc: 'trunc',
    arithmetic.TensorDegrees: 'degrees',
    arithmetic.TensorRadians: 'radians',

    arithmetic.TensorEqual: 'equal',
    arithmetic.TensorNotEqual: 'not_equal',
    arithmetic.TensorLessThan: 'less',
    arithmetic.TensorLessEqual: 'less_equal',
    arithmetic.TensorGreaterThan: 'greater',
    arithmetic.TensorGreaterEqual: 'greater_equal',

    arithmetic.TensorSin: 'sin',
    arithmetic.TensorCos: 'cos',
    arithmetic.TensorTan: 'tan',
    arithmetic.TensorArcsin: 'arcsin',
    arithmetic.TensorArccos: 'arccos',
    arithmetic.TensorArctan: 'arctan',
    arithmetic.TensorArctan2: 'arctan2',
    arithmetic.TensorHypot: 'hypot',
    arithmetic.TensorSinh: 'sinh',
    arithmetic.TensorCosh: 'cosh',
    arithmetic.TensorTanh: 'tanh',
    arithmetic.TensorArcsinh: 'arcsinh',
    arithmetic.TensorArccosh: 'arccosh',
    arithmetic.TensorArctanh: 'arctanh',
    arithmetic.TensorDeg2rad: 'deg2rad',
    arithmetic.TensorRad2deg: 'rad2deg',
    arithmetic.TensorAngle: 'angle',

    arithmetic.TensorBitand: 'bitwise_and',
    arithmetic.TensorBitor: 'bitwise_or',
    arithmetic.TensorBitxor: 'bitwise_xor',
    arithmetic.TensorInvert: 'invert',

    arithmetic.TensorLshift: 'left_shift',
    arithmetic.TensorRshift: 'right_shift',

    arithmetic.TensorAnd: 'logical_and',
    arithmetic.TensorOr: 'logical_or',
    arithmetic.TensorXor: 'logical_xor',
    arithmetic.TensorNot: 'logical_not',

    arithmetic.TensorMaximum: 'maximum',
    arithmetic.TensorMinimum: 'minimum',
    arithmetic.TensorFMax: 'fmax',
    arithmetic.TensorFMin: 'fmin',

    arithmetic.TensorIsclose: 'isclose',

    arithmetic.TensorClip: 'clip',
    arithmetic.TensorIsReal: 'isreal',
    arithmetic.TensorIsComplex: 'iscomplex',
    arithmetic.TensorReal: 'real',
    arithmetic.TensorImag: 'imag',
    arithmetic.TensorFix: 'fix',
    arithmetic.TensorI0: 'i0',
    arithmetic.TensorSinc: 'sinc',
    arithmetic.TensorNanToNum: 'nan_to_num',
}


def _handle_out_dtype(val, dtype):
    if val.dtype != dtype:
        return val.astype(dtype)
    return val


def _build_elementwise(op):
    def _handle(ctx, chunk):
        inputs, device_id, xp = as_same_device(
            [ctx[c.key] for c in chunk.inputs], device=chunk.device, ret_extra=True)

        if isinstance(op, six.string_types):
            func = getattr(xp, op)
        else:
            func = op

        with device(device_id):
            kw = {'casting': chunk.op.casting} if chunk.op.out else {}

            if chunk.op.out and chunk.op.where:
                inputs, kw['out'], kw['where'] = inputs[:-2], inputs[-2].copy(), inputs[-1]
            elif chunk.op.out:
                inputs, kw['out'] = inputs[:-1], inputs[-1].copy()

            with np.errstate(**chunk.op.err):
                if len(inputs) == 1:
                    try:
                        ctx[chunk.key] = _handle_out_dtype(func(inputs[0], **kw), chunk.op.dtype)
                    except TypeError:
                        if kw.get('where') is None:
                            raise
                        out, where = kw.pop('out'), kw.pop('where')
                        ctx[chunk.key] = _handle_out_dtype(xp.where(where, func(inputs[0]), out),
                                                           chunk.op.dtype)
                else:
                    try:
                        if is_sparse_module(xp):
                            ctx[chunk.key] = _handle_out_dtype(reduce(lambda a, b: func(a, b, **kw), inputs),
                                                               chunk.op.dtype)
                        else:
                            if 'out' not in kw:
                                dest_value = xp.empty(chunk.shape, chunk.dtype)
                                kw['out'] = dest_value
                            ctx[chunk.key] = _handle_out_dtype(reduce(lambda a, b: func(a, b, **kw), inputs),
                                                               chunk.op.dtype)
                    except TypeError:
                        if kw.get('where') is None:
                            raise
                        out, where = kw.pop('out'), kw.pop('where')
                        ctx[chunk.key] = _handle_out_dtype(
                            xp.where(where, reduce(lambda a, b: func(a, b), inputs), out),
                            chunk.op.dtype)
    return _handle


def _const_elementwise(op):
    def _handle(ctx, chunk):
        if chunk.inputs is not None:
            try:
                _, device_id, xp = as_same_device(
                    [ctx[c.key] for c in chunk.inputs], device=chunk.device, ret_extra=True)
            except KeyError:
                raise
        else:
            # all constants
            device_id, xp = -1, np

        if isinstance(op, six.string_types):
            func = getattr(xp, op)
        else:
            func = op

        get = lambda x: ctx[x.key] if hasattr(x, 'key') else x

        with device(device_id):
            with np.errstate(**chunk.op.err):
                ctx[chunk.key] = _handle_out_dtype(
                    reduce(func, (get(chunk.op.lhs), get(chunk.op.rhs))), chunk.op.dtype)
    return _handle


def _around(ctx, chunk):
    (a,), device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.device, ret_extra=True)

    with device(device_id):
        ctx[chunk.key] = xp.around(a, decimals=chunk.op.decimals)


def _angle(ctx, chunk):
    (z,), device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.device, ret_extra=True)

    with device(device_id):
        ctx[chunk.key] = xp.angle(z, deg=chunk.op.deg)


def _isclose(ctx, chunk):
    (a, b), device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.device, ret_extra=True)

    with device(device_id):
        ctx[chunk.key] = xp.isclose(a, b, atol=chunk.op.atol, rtol=chunk.op.rtol,
                                    equal_nan=chunk.op.equal_nan)


def _frexp(ctx, chunk):
    inputs, device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.device, ret_extra=True)

    with device(device_id):
        kw = {'casting': chunk.op.casting}

        inputs_iter = iter(inputs)
        input = next(inputs_iter)
        if chunk.op.out1 is not None:
            out1 = next(inputs_iter)
        else:
            out1 = None
        if chunk.op.out2 is not None:
            out2 = next(inputs_iter)
        else:
            out2 = None
        if chunk.op.where is not None:
            where = kw['where'] = next(inputs_iter)
        else:
            where = None

        try:
            args = [input]
            if out1 is not None:
                args.append(out1)
            if out2 is not None:
                args.append(out2)
            mantissa, exponent = xp.frexp(*args, **kw)
        except TypeError:
            if where is None:
                raise
            mantissa, exponent = xp.frexp(input)
            mantissa, exponent = xp.where(where, mantissa, out1), xp.where(where, exponent, out2)

        for c, res in zip(chunk.op.outputs, (mantissa, exponent)):
            ctx[c.key] = res


def _modf(ctx, chunk):
    inputs, device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.device, ret_extra=True)

    with device(device_id):
        kw = {'casting': chunk.op.casting}

        inputs_iter = iter(inputs)
        input = next(inputs_iter)
        if chunk.op.out1 is not None:
            out1 = next(inputs_iter)
        else:
            out1 = None
        if chunk.op.out2 is not None:
            out2 = next(inputs_iter)
        else:
            out2 = None
        if chunk.op.where is not None:
            where = kw['where'] = next(inputs_iter)
        else:
            where = None

        try:
            args = [input]
            if out1 is not None:
                args.append(out1.copy())
            if out2 is not None:
                args.append(out2.copy())
            y1, y2 = xp.modf(*args, **kw)
        except TypeError:
            if where is None:
                raise
            y1, y2 = xp.modf(input)
            y1, y2 = xp.where(where, y1, out1), xp.where(where, y2, out2)

        for c, res in zip(chunk.op.outputs, (y1, y2)):
            ctx[c.key] = res


def _set_real(ctx, chunk):
    inputs, device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.device, ret_extra=True)

    if len(inputs) == 1:
        val, real = inputs[0], chunk.op.rhs
    else:
        assert len(inputs) == 2
        val, real = inputs

    with device(device_id):
        val = val.copy()
        val.real = real

        ctx[chunk.key] = val


def _set_imag(ctx, chunk):
    inputs, device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.device, ret_extra=True)

    if len(inputs) == 1:
        val, imag = inputs[0], chunk.op.rhs
    else:
        assert len(inputs) == 2
        val, imag = inputs

    with device(device_id):
        val = val.copy()
        val.imag = imag

        ctx[chunk.key] = val


def _clip(ctx, chunk):
    inputs, device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.device, ret_extra=True)

    inputs_iter = iter(inputs)
    a = next(inputs_iter)
    a_min = next(inputs_iter) if isinstance(chunk.op.a_min, type(chunk)) else chunk.op.a_min
    a_max = next(inputs_iter) if isinstance(chunk.op.a_max, type(chunk)) else chunk.op.a_max
    out = next(inputs_iter).copy() if chunk.op.out is not None else None

    with device(device_id):
        kw = {}
        if out is not None:
            kw['out'] = out
        ctx[chunk.key] = xp.clip(a, a_min, a_max, **kw)


def _i0(ctx, chunk):
    x = ctx[chunk.inputs[0].key]
    xp = get_array_module(x)
    res = xp.i0(x)
    if not is_sparse_module(xp):
        res = res.reshape(chunk.shape)
    ctx[chunk.key] = res


def _tree_add(ctx, chunk):
    inputs, device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.device, ret_extra=True)

    with device(device_id):
        ctx[chunk.key] = reduce(xp.add, inputs)


def _tree_multiply(ctx, chunk):
    inputs, device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.device, ret_extra=True)

    with device(device_id):
        ctx[chunk.key] = reduce(xp.multiply, inputs)


def register_arithmetic_handler():
    from .core import register

    for op, new_op in six.iteritems(OP_TO_HANDLER):
        register(op, _build_elementwise(new_op))
        if hasattr(op, 'constant_cls'):
            const_op = op.constant_cls()
            if const_op:
                register(const_op, _const_elementwise(OP_TO_HANDLER[op]))

    register(arithmetic.TensorSetReal, _set_real)
    register(arithmetic.TensorSetRealConstant, _set_real)
    register(arithmetic.TensorSetImag, _set_imag)
    register(arithmetic.TensorSetImagConstant, _set_imag)

    register(arithmetic.TensorTreeAdd, _tree_add)
    register(arithmetic.TensorTreeMultiply, _tree_multiply)
    register(arithmetic.TensorAround, _around)
    register(arithmetic.TensorAngle, _angle)
    register(arithmetic.TensorIsclose, _isclose)
    register(arithmetic.TensorFrexp, _frexp)
    register(arithmetic.TensorModf, _modf)
    register(arithmetic.TensorClip, _clip)
    register(arithmetic.TensorI0, _i0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .expressions.datasource import tensor, array, asarray, scalar, \
    empty, empty_like, ones, ones_like, zeros, zeros_like, \
    full, arange, diag, diagflat, eye, identity, linspace, \
    meshgrid, indices, tril, triu
from .expressions.base import result_type, copyto, transpose, where, broadcast_to, broadcast_arrays, \
    expand_dims, rollaxis, swapaxes, moveaxis, ravel, atleast_1d, atleast_2d, atleast_3d, argwhere, \
    array_split, split, hsplit, vsplit, dsplit, roll, squeeze, ptp, diff, ediff1d, digitize, \
    average, cov, corrcoef, flip, flipud, fliplr, repeat, tile, isin
from .expressions.arithmetic import add, subtract, multiply, divide, truediv as true_divide, \
    floordiv as floor_divide, mod, power, float_power, fmod, sqrt, \
    around, round_, round_ as round, logaddexp, logaddexp2, negative, positive, \
    absolute, fabs, absolute as abs, rint, sign, degrees, radians, conj, conjugate, exp, exp2, \
    log, log2, log10, expm1, log1p, sqrt, square, cbrt, reciprocal, \
    equal, not_equal, less, less_equal, greater, greater_equal, sin, cos, tan, \
    arcsin, arccos, arctan, arctan2, hypot, sinh, cosh, tanh, arcsinh, arccosh, arctanh, \
    deg2rad, rad2deg, bitand as bitwise_and, bitor as bitwise_or, bitxor as bitwise_xor, \
    invert, invert as bitwise_not, lshift as left_shift, rshift as right_shift, \
    logical_and, logical_or, logical_xor, logical_not, \
    maximum, minimum, floor, ceil, trunc, remainder, fmax, fmin, isfinite, isinf, isnan, \
    signbit, copysign, nextafter, spacing, clip, isclose, ldexp, frexp, modf, angle, \
    isreal, iscomplex, real, imag, fix, i0, sinc, nan_to_num
from .expressions.linalg.tensordot import tensordot
from .expressions.linalg.dot import dot
from .expressions.linalg.inner import inner, innerproduct
from .expressions.linalg.vdot import vdot
from .expressions.linalg.matmul import matmul
from .expressions.reduction import sum, nansum, prod, prod as product, nanprod, \
    max, max as amax, nanmax, min, min as amin, nanmin, all, any, mean, nanmean, \
    argmax, nanargmax, argmin, nanargmin, cumsum, cumprod, \
    var, std, nanvar, nanstd, nancumsum, nancumprod, count_nonzero, allclose, array_equal
from .expressions.reshape import reshape
from .expressions.merge import concatenate, stack, hstack, vstack, dstack, column_stack
from .expressions.indexing import take, compress, extract, choose, unravel_index, nonzero, flatnonzero
from .expressions import random
from .expressions import fft
from .expressions import linalg
from .expressions import lib
from .expressions.lib.index_tricks import mgrid, ogrid, ndindex

from numpy import newaxis, AxisError, inf, Inf, NINF, nan, NAN, NaN, pi, e, \
    errstate, geterr, seterr
# import numpy types
from numpy import dtype, number, inexact, floating, complexfloating, \
    integer, signedinteger, unsignedinteger, character, generic, flexible, \
    int_, bool_, float_, cfloat, bytes_, unicode_, void, object_, \
    intc, intp, int8, int16, int32, int64, uint8, uint16, uint32, uint64, uint, \
    float16, float32, float64, complex64, complex128, datetime64, timedelta64
from numpy import finfo



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .compat import six, izip
from .utils import tokenize


class Base(object):
    __slots__ = ()
    _no_copy_attrs_ = set()

    def __init__(self, *args, **kwargs):
        for slot, arg in izip(self.__slots__, args):
            object.__setattr__(self, slot, arg)

        for key, val in six.iteritems(kwargs):
            object.__setattr__(self, key, val)

    @property
    def _values_(self):
        return [getattr(self, k, None) for k in self.__slots__
                if k not in self._no_copy_attrs_]


class BaseWithKey(Base):
    __slots__ = '_key', '_id'
    _no_copy_attrs_ = {'_id'}
    _init_update_key_ = True

    def __init__(self, *args, **kwargs):
        super(BaseWithKey, self).__init__(*args, **kwargs)

        if self._init_update_key_ and (not hasattr(self, '_key') or not self._key):
            self.update_key()
        if not hasattr(self, '_id') or not self._id:
            self._id = str(id(self))

    def _obj_set(self, k, v):
        object.__setattr__(self, k, v)

    def update_key(self):
        self._obj_set('_key', tokenize(type(self), *self._values_))
        return self

    def reset_key(self):
        self._obj_set('_key', None)
        return self

    def update_id(self, new_id=None):
        new_id = new_id if new_id is not None else str(id(self))
        self._obj_set('_id', new_id)

    def __copy__(self):
        return self.copy()

    def copy(self):
        return self.copy_to(type(self)(_key=self.key))

    def copy_to(self, target):
        for attr in self.__slots__:
            if (attr.startswith('__') and attr.endswith('__')) or attr in self._no_copy_attrs_:
                # we don't copy id to identify that the copied one is new
                continue
            if hasattr(self, attr):
                setattr(target, attr, getattr(self, attr))

        return target

    def copy_from(self, obj):
        obj.copy_to(self)

    @property
    def key(self):
        return self._key

    @property
    def id(self):
        return self._id


class Entity(object):
    __slots__ = '_data',
    _allow_data_type_ = ()

    def __init__(self, data):
        self._check_data(data)
        self._data = data

    def _check_data(self, data):
        if data is not None and not isinstance(data, self._allow_data_type_):
            raise TypeError('Expect {0}, got {1}'.format(self._allow_data_type_, type(data)))

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, new_data):
        self._check_data(new_data)
        self._data = new_data

    def __copy__(self):
        return self.copy()

    def copy(self):
        self.copy_to(type(self)(None))

    def copy_to(self, target):
        target.data = self._data

    def copy_from(self, obj):
        self.data = obj.data

    def __getattr__(self, attr):
        return getattr(self._data, attr)

    def __setattr__(self, key, value):
        try:
            super(Entity, self).__setattr__(key, value)
        except AttributeError:
            return setattr(self._data, key, value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .sum import TensorSum
from .core import TensorReduction


class TensorCountNonzero(operands.CountNonzero, TensorReduction):
    def __init__(self, axis=None, dtype=np.intp, keepdims=None, combine_size=None, **kw):
        super(TensorCountNonzero, self).__init__(_axis=axis, _dtype=dtype, _keepdims=keepdims,
                                                 _combine_size=combine_size, **kw)

    @staticmethod
    def _get_op_types():
        return TensorCountNonzero, TensorSum, None


def count_nonzero(a, axis=None, combine_size=None):
    """
    Counts the number of non-zero values in the tensor ``a``.

    The word "non-zero" is in reference to the Python 2.x
    built-in method ``__nonzero__()`` (renamed ``__bool__()``
    in Python 3.x) of Python objects that tests an object's
    "truthfulness". For example, any number is considered
    truthful if it is nonzero, whereas any string is considered
    truthful if it is not the empty string. Thus, this function
    (recursively) counts how many elements in ``a`` (and in
    sub-tensors thereof) have their ``__nonzero__()`` or ``__bool__()``
    method evaluated to ``True``.

    Parameters
    ----------
    a : array_like
        The tensor for which to count non-zeros.
    axis : int or tuple, optional
        Axis or tuple of axes along which to count non-zeros.
        Default is None, meaning that non-zeros will be counted
        along a flattened version of ``a``.
    combine_size: int, optional
        The number of chunks to combine.

    Returns
    -------
    count : int or tensor of int
        Number of non-zero values in the array along a given axis.
        Otherwise, the total number of non-zero values in the tensor
        is returned.

    See Also
    --------
    nonzero : Return the coordinates of all the non-zero values.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.count_nonzero(mt.eye(4)).execute()
    4
    >>> mt.count_nonzero([[0,1,7,0,0],[3,0,0,2,19]]).execute()
    5
    >>> mt.count_nonzero([[0,1,7,0,0],[3,0,0,2,19]], axis=0).execute()
    array([1, 1, 1, 1, 1])
    >>> mt.count_nonzero([[0,1,7,0,0],[3,0,0,2,19]], axis=1).execute()
    array([2, 3])

    """
    op = TensorCountNonzero(axis=axis, dtype=np.intp, keepdims=None, combine_size=combine_size)
    return op(a)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np

from ....operands import Astype
from ..core import TensorOperandMixin


class TensorAstype(Astype, TensorOperandMixin):
    def __init__(self, dtype=None, casting=None, sparse=False, **kw):
        super(TensorAstype, self).__init__(_dtype=dtype, _casting=casting,
                                           _sparse=sparse, **kw)

    def _set_inputs(self, inputs):
        super(TensorAstype, self)._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, tensor, copy=True):
        t = self.new_tensor([tensor], tensor.shape)
        if copy:
            return t
        tensor.data = t.data
        return tensor

    @classmethod
    def tile(cls, op):
        in_tensor = op.input

        out_chunks = []
        for c in in_tensor.chunks:
            chunk_op = op.copy().reset_key()
            chunk = chunk_op.new_chunk([c], c.shape, index=c.index)
            out_chunks.append(chunk)

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, op.outputs[0].shape, nsplits=in_tensor.nsplits,
                                  chunks=out_chunks)


def _astype(tensor, dtype, casting='unsafe', copy=True):
    """
    Copy of the tensor, cast to a specified type.

    Parameters
    ----------
    dtype : str or dtype
        Typecode or data-type to which the array is cast.
    casting : {'no', 'equiv', 'safe', 'same_kind', 'unsafe'}, optional
        Controls what kind of data casting may occur. Defaults to 'unsafe'
        for backwards compatibility.

          * 'no' means the data types should not be cast at all.
          * 'equiv' means only byte-order changes are allowed.
          * 'safe' means only casts which can preserve values are allowed.
          * 'same_kind' means only safe casts or casts within a kind,
            like float64 to float32, are allowed.
          * 'unsafe' means any data conversions may be done.
    copy : bool, optional
        By default, astype always returns a newly allocated array. If this
        is set to false, and the `dtype`, `order`, and `subok`
        requirements are satisfied, the input array is returned instead
        of a copy.

    Returns
    -------
    arr_t : Tensor
        Unless `copy` is False and the other conditions for returning the input
        array are satisfied (see description for `copy` input parameter), `arr_t`
        is a new tensor of the same shape as the input array, with dtype, order
        given by `dtype`, `order`.

    Notes
    -----
    astype method returns an error if the string
    dtype to cast to is not long enough in 'safe' casting mode to hold the max
    value of integer/float array that is being casted. Previously the casting
    was allowed even if the result was truncated.

    Raises
    ------
    ComplexWarning
        When casting from complex to float or int. To avoid this,
        one should use ``a.real.astype(t)``.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> x = mt.array([1, 2, 2.5])
    >>> x.execute()
    array([ 1. ,  2. ,  2.5])

    >>> x.astype(int).execute()
    array([1, 2, 2])
    """
    dtype = np.dtype(dtype)
    if tensor.dtype == dtype:
        return tensor
    elif not np.can_cast(tensor.dtype, dtype, casting=casting):
        raise TypeError('Cannot cast array from {0!r} to {1!r} '
                        'according to the rule {2!s}'.format(
            tensor.dtype, dtype, casting))

    op = TensorAstype(dtype=dtype, casting=casting, sparse=tensor.issparse())
    return op(tensor, copy=copy)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .cholesky import cholesky
from .dot import dot
from .inner import inner, innerproduct
from .inv import inv
from .lu import lu
from .matmul import matmul
from .norm import norm
from .qr import qr
from .solve import solve
from .solve_triangular import solve_triangular
from .svd import svd
from .tensordot import tensordot
from .vdot import vdot


def _install():
    from ...core import Tensor, TensorData

    setattr(Tensor, '__matmul__', matmul)
    setattr(TensorData, '__matmul__', matmul)
    setattr(TensorData, 'dot', dot)


_install()
del _install



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

from .utils import SchedulerActor
from .graph import GraphActor
from ..compat import six

logger = logging.getLogger(__name__)


class SessionActor(SchedulerActor):
    def __init__(self, session_id, **kwargs):
        super(SessionActor, self).__init__()
        self._session_id = session_id
        self._args = kwargs

        self._cluster_info_ref = None
        self._assigner_ref = None
        self._graph_refs = dict()

    @staticmethod
    def gen_name(session_id):
        return 's:session$%s' % session_id

    def get_argument(self, key):
        return self._args[key]

    def get_graph_refs(self):
        return self._graph_refs

    def post_create(self):
        logger.debug('Actor %s running in process %d', self.uid, os.getpid())
        self.set_cluster_info_ref()

        from .assigner import AssignerActor
        assigner_uid = AssignerActor.gen_name(self._session_id)
        assigner_addr = self.get_scheduler(assigner_uid)
        self._assigner_ref = self.ctx.create_actor(
            AssignerActor, uid=assigner_uid, address=assigner_addr)

    def pre_destroy(self):
        self.ctx.destroy_actor(self._assigner_ref)
        [self.ctx.destroy_actor(graph_ref) for graph_ref in six.itervalues(self._graph_refs)]

    def submit_tensor_graph(self, serialized_graph, graph_key, target_tensors=None):
        graph_uid = GraphActor.gen_name(self._session_id, graph_key)
        graph_ref = self.ctx.create_actor(GraphActor, self._session_id, graph_key,
                                          serialized_graph, target_tensors=target_tensors,
                                          uid=graph_uid, address=self.get_scheduler(graph_uid))
        graph_ref.execute_graph(_tell=True)
        self._graph_refs[graph_key] = graph_ref

    def graph_state(self, graph_key):
        return self._graph_refs[graph_key].get_state()

    def fetch_result(self, graph_key, tensor_key):
        # TODO just for test, should move to web handler
        graph_ref = self._graph_refs[graph_key]
        return graph_ref.fetch_tensor_result(tensor_key)


class SessionManagerActor(SchedulerActor):
    def __init__(self):
        super(SessionManagerActor, self).__init__()
        self._sessions = dict()

    @classmethod
    def default_name(cls):
        return 's:%s' % cls.__name__

    def post_create(self):
        logger.debug('Actor %s running in process %d', self.uid, os.getpid())

        self.set_cluster_info_ref()

    def get_sessions(self):
        return self._sessions

    def create_session(self, session_id, **kwargs):
        uid = SessionActor.gen_name(session_id)
        scheduler_address = self.get_scheduler(uid)
        session_ref = self.ctx.create_actor(SessionActor, uid=uid, address=scheduler_address,
                                            session_id=session_id, **kwargs)
        self._sessions[session_id] = session_ref
        return session_ref

    def delete_session(self, session_id):
        if session_id in self._sessions:
            session_ref = self._sessions[session_id]
            session_ref.destroy()
            del self._sessions[session_id]
